#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK

import argcomplete
import argparse
import math
import multiprocessing
import tarfile
import time
from pathlib import Path
import artistools as at
import numpy as np
import pandas as pd
from functools import lru_cache, partial

traj_root = at.config['gsimerger_trajroot']


def get_elemabund_from_nucabund(dfnucabund):
    dictelemabund = {}
    for atomic_number in range(1, dfnucabund.Z.max() + 1):
        dictelemabund[f'X_{at.get_elsymbol(atomic_number)}'] = (
            dfnucabund.query('Z == @atomic_number', inplace=False).massfrac.sum())
    return dictelemabund


def get_traj_tarpath(particleid):
    return Path(traj_root, f'{particleid}.tar.xz')


def open_tar_file_or_extracted(particleid, memberfilename):
    """ memberfilename is within the trajectory tarfile/folder, eg. ./Run_rprocess/evol.dat """
    path_extracted_file = Path(traj_root, str(particleid), memberfilename)
    if path_extracted_file.is_file():
        return open(path_extracted_file, mode='r')
    else:
        return tarfile.open(get_traj_tarpath(particleid), mode='r:*').extractfile(member=memberfilename)


@lru_cache(maxsize=16)
def get_dfevol(particleid):
    with open_tar_file_or_extracted(particleid, './Run_rprocess/evol.dat') as evolfile:
        dfevol = pd.read_csv(
            evolfile, delim_whitespace=True, comment='#', usecols=[0, 1], names=['nstep', 'timesec'])
    return dfevol


def get_closest_network_timestep(particleid, timesec, cond='nearest'):
    """
        cond:
            'lessthan': find highest timestep less than time_sec
            'greaterthan': find lowest timestep greater than time_sec
    """
    dfevol = get_dfevol(particleid)
    if cond == 'nearest':
        idx = np.abs(dfevol.timesec.values - timesec).argmin()
    elif cond == 'greaterthan':
        return dfevol.query('timesec > @timesec').nstep.min()
    elif cond == 'lessthan':
        return dfevol.query('timesec < @timesec').nstep.max()
    else:
        assert False
    nts = dfevol.nstep.values[idx]

    return nts


def get_trajectory_timestepfile_nuc_abund(particleid, memberfilename):
    with open_tar_file_or_extracted(particleid, memberfilename) as trajfile:

        # with open(trajfile) as ftraj:
        _, str_t_model_init_seconds, _, rho, _, _ = trajfile.readline().split()
        t_model_init_seconds = float(str_t_model_init_seconds)
        dfnucabund = pd.read_csv(trajfile, delim_whitespace=True, comment='#',
                                 names=["N", "Z", "log10abund", "S1n", "S2n"], usecols=["N", "Z", "log10abund"],
                                 dtype={0: int, 1: int, 2: float})

    # dfnucabund.eval('abund = 10 ** log10abund', inplace=True)
    dfnucabund.eval('massfrac = (N + Z) * (10 ** log10abund)', inplace=True)
    # dfnucabund.eval('A = N + Z', inplace=True)
    # dfnucabund.query('abund > 0.', inplace=True)

    # abund is proportional to number abundance, but needs normalisation
    # normfactor = dfnucabund.abund.sum()
    # print(f'abund sum: {normfactor}')
    # dfnucabund.eval('numberfrac = abund / @normfactor', inplace=True)

    return dfnucabund, t_model_init_seconds


def get_trajectory_nuc_abund(particleid, t_model_s=None, nts=None):
    """
        nts: GSI network timestep number
    """
    assert t_model_s is not None or nts is not None
    try:
        if nts is not None:
            memberfilename = f'./Run_rprocess/nz-plane{nts:05d}'
        else:
            if np.isclose(t_model_s, 86400, rtol=0.1):
                memberfilename = './Run_rprocess/tday_nz-plane'
            else:
                # find the closest timestep to the required time
                nts = get_closest_network_timestep(particleid, t_model_s)
                memberfilename = f'./Run_rprocess/nz-plane{nts:05d}'

        dftrajnucabund, traj_time_s = get_trajectory_timestepfile_nuc_abund(particleid, memberfilename)

    except FileNotFoundError:
        # print(f' WARNING {get_traj_tarpath(particleid)} not found! ')
        return None

    massfractotal = dftrajnucabund.massfrac.sum()
    dftrajnucabund.query('Z >= 1', inplace=True)

    dftrajnucabund['nucabundcolname'] = [f'X_{at.get_elsymbol(int(row.Z))}{int(row.N + row.Z)}'
                                         for row in dftrajnucabund.itertuples()]

    colmassfracs = list(dftrajnucabund[['nucabundcolname', 'massfrac']].itertuples(index=False))
    colmassfracs.sort(key=lambda row: at.get_z_a_nucname(row[0]))

    # print(f'trajectory particle id {particleid} massfrac sum: {massfractotal:.2f}')
    # print(f' grid snapshot: {t_model_s:.2e} s, network: {traj_time_s:.2e} s (timestep {nts})')
    assert np.isclose(massfractotal, 1., rtol=0.02)
    if nts is None:
        assert np.isclose(traj_time_s, t_model_s, rtol=0.2, atol=1.)

    dict_traj_nuc_abund = {nucabundcolname: massfrac / massfractotal for nucabundcolname, massfrac in colmassfracs}
    return dict_traj_nuc_abund


def get_modelcellabundance(dict_traj_nuc_abund, minparticlespercell, cellgroup):
    cellindex, dfthiscellcontribs = cellgroup

    if len(dfthiscellcontribs) < minparticlespercell:
        return None

    contribparticles = [
        (dict_traj_nuc_abund[particleid], frac_of_cellmass)
        for particleid, frac_of_cellmass in dfthiscellcontribs[
            ['particleid', 'frac_of_cellmass']].itertuples(index=False)
        if particleid in dict_traj_nuc_abund]

    # adjust frac_of_cellmass for missing particles
    cell_frac_sum = sum([frac_of_cellmass for _, frac_of_cellmass in contribparticles])

    nucabundcolnames = set([
        col for particleid in dfthiscellcontribs.particleid
        for col in dict_traj_nuc_abund.get(particleid, {}).keys()])

    row = {
        nucabundcolname: sum([
            frac_of_cellmass * traj_nuc_abund.get(nucabundcolname, 0.) / cell_frac_sum
            for traj_nuc_abund, frac_of_cellmass in contribparticles]
        ) for nucabundcolname in nucabundcolnames}
    row['inputcellid'] = cellindex

    # if n % 100 == 0:
    #     functime = time.perf_counter() - timefuncstart
    #     print(f'cell id {cellindex:6d} ('
    #           f'{n:4d} of {active_inputcellcount:4d}, {n / active_inputcellcount * 100:4.1f}%) '
    #           f' contributing {len(dfthiscellcontribs):4d} particles.'
    #           f' total func time {functime:.1f} s, {n / functime:.1f} cell/s,'
    #           f' expected time: {functime / n * active_inputcellcount:.1f}')
    return row


def get_gridparticlecontributions(gridcontribpath):
    dfcontribs = pd.read_csv(
        Path(gridcontribpath, 'gridcontributions.txt'), delim_whitespace=True,
        dtype={0: int, 1: int, 2: float, 3: float})

    return dfcontribs


def filtermissinggridparticlecontributions(dfcontribs):
    missing_particleids = []
    for particleid in sorted(dfcontribs.particleid.unique()):
        if not get_traj_tarpath(particleid).is_file():
            missing_particleids.append(particleid)
            # print(f' WARNING particle {particleid} not found!')

    print(f'Adding column that excludes contributions from {len(missing_particleids)} '
          'particles without abundance data and renormalising...', end='')
    # after filtering, frac_of_cellmass_includemissing will still include particles with rho but no abundance data
    # frac_of_cellmass will exclude particles with no abundances
    dfcontribs.eval('frac_of_cellmass_includemissing = frac_of_cellmass', inplace=True)
    # dfcontribs.query('particleid not in @missing_particleids', inplace=True)
    dfcontribs.loc[dfcontribs.eval('particleid in @missing_particleids'), 'frac_of_cellmass'] = 0.

    dfcontribs['frac_of_cellmass'] = [
        row.frac_of_cellmass if row.particleid not in missing_particleids else 0. for row in dfcontribs.itertuples()]

    cell_frac_sum = {}
    cell_frac_includemissing_sum = {}
    for cellindex, dfparticlecontribs in dfcontribs.groupby('cellindex'):
        cell_frac_sum[cellindex] = dfparticlecontribs.frac_of_cellmass.sum()
        cell_frac_includemissing_sum[cellindex] = dfparticlecontribs.frac_of_cellmass_includemissing.sum()

    dfcontribs['frac_of_cellmass'] = [
        row.frac_of_cellmass / cell_frac_sum[row.cellindex]
        if cell_frac_sum[row.cellindex] > 0. else 0.
        for row in dfcontribs.itertuples()]

    dfcontribs['frac_of_cellmass_includemissing'] = [
        row.frac_of_cellmass_includemissing / cell_frac_includemissing_sum[row.cellindex]
        if cell_frac_includemissing_sum[row.cellindex] > 0. else 0.
        for row in dfcontribs.itertuples()]

    for cellindex, dfparticlecontribs in dfcontribs.groupby('cellindex'):
        frac_sum = dfparticlecontribs.frac_of_cellmass.sum()
        assert frac_sum == 0. or np.isclose(frac_sum, 1., rtol=0.02)

        cell_frac_includemissing_sum = dfparticlecontribs.frac_of_cellmass_includemissing.sum()
        assert cell_frac_includemissing_sum == 0. or np.isclose(cell_frac_includemissing_sum, 1., rtol=0.02)

    print('done')

    return dfcontribs


def save_gridparticlecontributions(dfcontribs, gridcontribpath):
    gridcontribpath = Path(gridcontribpath)
    if gridcontribpath.is_dir():
        gridcontribpath = Path(gridcontribpath, 'gridcontributions.txt')
    dfcontribs.to_csv(gridcontribpath, sep=' ', index=False)


def add_abundancecontributions(dfgridcontributions, dfmodel, t_model_days, minparticlespercell=0):
    """ contribute trajectory network calculation abundances to model cell abundances """
    t_model_s = t_model_days * 86400
    dfcontribs = dfgridcontributions

    if 'X_Fegroup' not in dfmodel.columns:
        dfmodel = pd.concat([dfmodel, pd.DataFrame({'X_Fegroup': np.ones(len(dfmodel))})], axis=1)

    active_inputcellids = [
        cellindex for cellindex, dfthiscellcontribs in dfcontribs.groupby('cellindex')
        if len(dfthiscellcontribs) >= minparticlespercell]

    dfcontribs.query('cellindex in @active_inputcellids', inplace=True)
    dfcontribs = filtermissinggridparticlecontributions(dfcontribs)
    active_inputcellids = dfcontribs.cellindex.unique()
    active_inputcellcount = len(active_inputcellids)

    dfcontribs_particlegroups = dfcontribs.groupby('particleid')
    particle_count = len(dfcontribs_particlegroups)

    print(f'{active_inputcellcount} of {len(dfmodel)} model cells have >={minparticlespercell} particles contributing '
          f'({len(dfcontribs)} total contributions from {particle_count} particles after filter)')

    listcellnucabundances = []
    print('Reading trajectory abundances...')
    timestart = time.perf_counter()
    trajnucabundworker = partial(get_trajectory_nuc_abund, t_model_s=t_model_s)

    if at.num_processes > 1:
        with multiprocessing.Pool(processes=at.num_processes) as pool:
            list_traj_nuc_abund = pool.map(trajnucabundworker, dfcontribs_particlegroups.groups)
            pool.close()
            pool.join()
    else:
        list_traj_nuc_abund = [trajnucabundworker(particleid) for particleid in dfcontribs_particlegroups.groups]

    n_missing_particles = len([d for d in list_traj_nuc_abund if d is None])
    print(f'  {n_missing_particles} particles are missing network abundance data')
    dict_traj_nuc_abund = {
        particleid: dftrajnucabund
        for particleid, dftrajnucabund in zip(dfcontribs_particlegroups.groups, list_traj_nuc_abund)
        if dftrajnucabund is not None}
    print(f'Reading trajectory abundances took {time.perf_counter() - timestart:.1f} seconds')

    print('Generating cell abundances...')
    timestart = time.perf_counter()
    dfcontribs_cellgroups = dfcontribs.groupby('cellindex')
    cellabundworker = partial(get_modelcellabundance, dict_traj_nuc_abund, minparticlespercell)

    if at.num_processes > 1:
        chunksize = math.ceil(len(dfcontribs_cellgroups) / at.num_processes)
        with multiprocessing.Pool(processes=at.num_processes) as pool:
            listcellnucabundances = pool.map(cellabundworker, dfcontribs_cellgroups, chunksize=chunksize)
            pool.close()
            pool.join()
    else:
        listcellnucabundances = [cellabundworker(cellgroup) for cellgroup in dfcontribs_cellgroups]

    listcellnucabundances = [x for x in listcellnucabundances if x is not None]
    print(f'  took {time.perf_counter() - timestart:.1f} seconds')

    timestart = time.perf_counter()
    print('Creating dfnucabundances...', end='', flush=True)
    dfnucabundances = pd.DataFrame(listcellnucabundances)
    dfnucabundances.set_index('inputcellid', drop=False, inplace=True)
    dfnucabundances.index.name = None
    dfnucabundances.fillna(0., inplace=True)
    print(f' took {time.perf_counter() - timestart:.1f} seconds')

    timestart = time.perf_counter()
    print('Adding up isotopes for elemental abundances and creating dfelabundances...', end='', flush=True)
    elemisotopes = {}
    for colname in sorted(dfnucabundances.columns):
        if not colname.startswith('X_'):
            continue
        atomic_number = at.get_atomic_number(colname[2:].rstrip('0123456789'))
        if atomic_number in elemisotopes:
            elemisotopes[atomic_number].append(colname)
        else:
            elemisotopes[atomic_number] = [colname]

    dfelabundances_partial = pd.DataFrame({
        'inputcellid': dfnucabundances.inputcellid,
        **{f'X_{at.get_elsymbol(atomic_number)}': dfnucabundances.eval(f'{" + ".join(elemisotopes[atomic_number])}')
            if atomic_number in elemisotopes else np.zeros(len(dfnucabundances))
            for atomic_number in range(1, max(elemisotopes.keys()) + 1)}}, index=dfnucabundances.index)

    # ensure cells with no traj contributions are included
    dfelabundances = dfmodel[['inputcellid']].merge(
        dfelabundances_partial, how='left', left_on='inputcellid', right_on='inputcellid')
    dfnucabundances.set_index('inputcellid', drop=False, inplace=True)
    dfnucabundances.index.name = None
    dfelabundances.fillna(0., inplace=True)
    print(f' took {time.perf_counter() - timestart:.1f} seconds')

    timestart = time.perf_counter()
    print('Merging isotopic abundances into dfmodel...', end='', flush=True)
    dfmodel = dfmodel.merge(dfnucabundances, how='left', left_on='inputcellid', right_on='inputcellid')
    dfmodel.fillna(0., inplace=True)
    print(f' took {time.perf_counter() - timestart:.1f} seconds')

    return dfmodel, dfelabundances, dfcontribs


def addargs(parser):
    parser.add_argument('-outputpath', '-o',
                        default='.',
                        help='Path for output files')


def main(args=None, argsraw=None, **kwargs):
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Create solar r-process pattern in ARTIS format.')

        addargs(parser)
        parser.set_defaults(**kwargs)
        argcomplete.autocomplete(parser)
        args = parser.parse_args(argsraw)

    # particleid = 88969  # Ye = 0.0963284224
    particleid = 133371  # Ye = 0.403913230
    print(f'trajectory particle id {particleid}')
    dfnucabund, t_model_init_seconds = get_trajectory_timestepfile_nuc_abund(particleid, './Run_rprocess/tday_nz-plane')
    dfnucabund.query('Z >= 1', inplace=True)
    dfnucabund['radioactive'] = True

    t_model_init_days = t_model_init_seconds / (24 * 60 * 60)

    wollager_profilename = 'wollager_ejectaprofile_10bins.txt'
    if Path(wollager_profilename).exists():
        print(f'{wollager_profilename} found')
        t_model_init_days_in = float(Path(wollager_profilename).open('rt').readline().strip().removesuffix(' day'))
        dfdensities = pd.read_csv(wollager_profilename, delim_whitespace=True, skiprows=1,
                                  names=['cellid', 'velocity_outer', 'rho'])
        dfdensities['cellid'] = dfdensities['cellid'].astype(int)
        dfdensities['velocity_inner'] = np.concatenate(([0.], dfdensities['velocity_outer'].values[:-1]))

        t_model_init_seconds_in = t_model_init_days_in * 24 * 60 * 60
        dfdensities.eval('cellmass_grams = rho * 4. / 3. * @math.pi * (velocity_outer ** 3 - velocity_inner ** 3)'
                         '* (1e5 * @t_model_init_seconds_in) ** 3', inplace=True)

        # now replace the density at the input time with the density at required time

        dfdensities.eval('rho = cellmass_grams / ('
                         '4. / 3. * @math.pi * (velocity_outer ** 3 - velocity_inner ** 3)'
                         ' * (1e5 * @t_model_init_seconds) ** 3)', inplace=True)
    else:
        rho = 1e-11
        print(f'{wollager_profilename} not found. Using rho {rho} g/cm3')
        dfdensities = pd.DataFrame(dict(rho=rho, velocity_outer=6.e4), index=[0])

    # print(dfdensities)

    # write abundances.txt
    dictelemabund = get_elemabund_from_nucabund(dfnucabund)

    dfelabundances = pd.DataFrame([dict(inputcellid=mgi + 1, **dictelemabund) for mgi in range(len(dfdensities))])
    # print(dfelabundances)
    at.inputmodel.save_initialabundances(dfelabundances=dfelabundances, abundancefilename=args.outputpath)

    # write model.txt

    rowdict = {
        # 'inputcellid': 1,
        # 'velocity_outer': 6.e4,
        # 'logrho': -3.,
        'X_Fegroup': 1.,
        'X_Ni56': 0.,
        'X_Co56': 0.,
        'X_Fe52': 0.,
        'X_Cr48': 0.,
        'X_Ni57': 0.,
        'X_Co57': 0.,
    }

    for _, row in dfnucabund.query('radioactive == True').iterrows():
        A = row.N + row.Z
        rowdict[f'X_{at.get_elsymbol(row.Z)}{A}'] = row.massfrac

    modeldata = []
    for mgi, densityrow in dfdensities.iterrows():
        # print(mgi, densityrow)
        modeldata.append(dict(inputcellid=mgi + 1, velocity_outer=densityrow['velocity_outer'],
                         logrho=math.log10(densityrow['rho']), **rowdict))
    # print(modeldata)

    dfmodel = pd.DataFrame(modeldata)
    # print(dfmodel)
    at.inputmodel.save_modeldata(dfmodel=dfmodel, t_model_init_days=t_model_init_days, modelpath=Path(args.outputpath))
    with open(Path(args.outputpath, 'gridcontributions.txt'), 'w') as fcontribs:
        fcontribs.write('particleid cellindex frac_of_cellmass\n')
        for cell in dfmodel.itertuples(index=False):
            fcontribs.write(f'{particleid} {cell.inputcellid} {1.}\n')


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import artistools as at
# import artistools.spectra
# import artistools.lightcurve.writebollightcurvedata

from pathlib import Path
import pandas as pd
import matplotlib.pyplot as plt
import os


def plot_hesma_spectrum(timeavg, axes):

    hesma_file = Path("/Users/ccollins/Downloads/hesma_files/M2a/hesma_specseq.dat")
    hesma_spec = pd.read_csv(hesma_file, comment="#", delim_whitespace=True, dtype=float)
    # print(hesma_spec)

    def match_closest_time(reftime):
        return str("{}".format(min([float(x) for x in hesma_spec.keys()[1:]], key=lambda x: abs(x - reftime))))

    closest_time = (match_closest_time(timeavg))
    closest_time = f'{closest_time:.2f}'
    print(closest_time)

    # Scale distance to 1 Mpc
    dist_mpc = 1e-5  # HESMA specta at 10 pc
    hesma_spec[closest_time] = hesma_spec[closest_time] * (1e-5) ** 2  # refspecditance Mpc / 1 Mpc ** 2

    for ax in axes:
        ax.plot(hesma_spec['0.00'], hesma_spec[closest_time], label='HESMA model')


def plothesmaresspec(fig, ax):

    # specfiles = ["/Users/ccollins/Downloads/hesma_files/M2a_i55/hesma_specseq_theta.dat"]
    specfiles = ["/Users/ccollins/Downloads/hesma_files/M2a/hesma_virtualspecseq_theta.dat"]

    for specfilename in specfiles:
        specdata = pd.read_csv(specfilename, delim_whitespace=True, header=None, dtype=float)

        # index_to_split = specdata.index[specdata.iloc[:, 1] == specdata.iloc[0, 1]]
        # res_specdata = []
        # for i, index_value in enumerate(index_to_split):
        #     if index_value != index_to_split[-1]:
        #         chunk = specdata.iloc[index_to_split[i]:index_to_split[i + 1], :]
        #     else:
        #         chunk = specdata.iloc[index_to_split[i]:, :]
        #     res_specdata.append(chunk)

        res_specdata = at.gather_res_data(specdata)

        column_names = res_specdata[0].iloc[0]
        column_names[0] = 'lambda'
        print(column_names)

        for i, res_spec in enumerate(res_specdata):
            res_specdata[i] = res_specdata[i].rename(columns=column_names).drop(res_specdata[i].index[0])

        ax.plot(res_specdata[0]['lambda'], res_specdata[0][11.7935] * (1e-5) ** 2, label="hesma 0")
        ax.plot(res_specdata[1]['lambda'], res_specdata[1][11.7935] * (1e-5) ** 2, label="hesma 1")
        ax.plot(res_specdata[2]['lambda'], res_specdata[2][11.7935] * (1e-5) ** 2, label="hesma 2")
        ax.plot(res_specdata[3]['lambda'], res_specdata[3][11.7935] * (1e-5) ** 2, label="hesma 3")
        ax.plot(res_specdata[4]['lambda'], res_specdata[4][11.7935] * (1e-5) ** 2, label="hesma 4")

    fig.legend()
    # plt.show()


def make_hesma_vspecfiles(modelpath):
    angles = [0, 1, 2, 3, 4]

    for angle in angles:
        vspecdata_all = at.spectra.get_specpol_data(angle=angle, modelpath=modelpath)
        vspecdata = vspecdata_all['I']

        timearray = vspecdata.columns.values[1:]
        vspecdata.sort_values(by='nu', ascending=False, inplace=True)
        vspecdata.eval('lambda_angstroms = 2.99792458e+18 / nu', inplace=True)
        for time in timearray:
            vspecdata[time] = vspecdata[time] * vspecdata['nu'] / vspecdata['lambda_angstroms']
            vspecdata[time] = vspecdata[time] * (1e5) ** 2  # Scale to 10 pc (1 Mpc/10 pc) ** 2

        vspecdata = vspecdata.set_index('lambda_angstroms').reset_index()
        vspecdata = vspecdata.drop(['nu'], axis=1)

        vspecdata = vspecdata.rename(columns={'lambda_angstroms': '0'})

        print(vspecdata)

        if angle == 0:
            vspecdata.to_csv(modelpath / 'hesma_virtualspecseq_theta.dat', sep=' ', index=False)  # create file
        else:
            # append to file
            vspecdata.to_csv(modelpath / 'hesma_virtualspecseq_theta.dat', mode='a', sep=' ', index=False)


def make_hesma_bol_lightcurve(modelpath, outpath, timemin, timemax):
    """UVOIR bolometric light curve (angle-averaged)"""

    lightcurvedataframe = at.lightcurve.writebollightcurvedata.get_bol_lc_from_lightcurveout(modelpath)
    print(lightcurvedataframe)
    lightcurvedataframe = lightcurvedataframe[lightcurvedataframe.time > timemin]
    lightcurvedataframe = lightcurvedataframe[lightcurvedataframe.time < timemax]

    modelname = at.get_model_name(modelpath)
    outfilename = f'doubledet_2021_{modelname}.dat'

    lightcurvedataframe.to_csv(outpath / outfilename, sep=' ', index=False, header=False)


def make_hesma_peakmag_dm15_dm40(pathtofiles, modelname, outpath):

    dm15filename = f"bolband_{modelname}_viewing_angle_data.txt"
    dm40filename = f"bolband_{modelname}_viewing_angle_data_deltam40.txt"

    dm15data = pd.read_csv(pathtofiles / dm15filename, delim_whitespace=True, header=None, names=['peakmag', 'risetime', 'dm15'], skiprows=1)
    dm40data = pd.read_csv(pathtofiles / dm40filename, delim_whitespace=True, header=None, names=['peakmag', 'risetime', 'dm40'], skiprows=1)

    outdata = {}
    outdata['peakmag'] = dm15data['peakmag']  # dm15 peak mag probably more accurate - shorter time window
    outdata['dm15'] = dm15data['dm15']
    outdata['dm40'] = dm40data['dm40']

    outdataframe = pd.DataFrame(outdata)
    outdataframe = outdataframe.round(decimals=4)
    outdataframe.to_csv(outpath / f"doubledet_2021_bol_dm_{modelname}.dat", sep=' ', index=False, header=True)


def read_hesma_peakmag_dm15_dm40(pathtofiles):
    data = []
    for filename in os.listdir(pathtofiles):
        print(filename)
        data.append(pd.read_csv(pathtofiles / filename, delim_whitespace=True))
    print(data[0])

    for df in data:
        print(df)
        plt.scatter(df['dm15'], df['peakmag'])
    plt.gca().invert_yaxis()
    plt.show()


# def main():
#     # pathtomodel = Path("/home/localadmin_ccollins/harddrive4TB/parameterstudy/")
#     # modelnames = ['M08_03', 'M08_05', 'M08_10', 'M09_03', 'M09_05', 'M09_10',
#     #               'M10_02_end55', 'M10_03', 'M10_05', 'M10_10', 'M11_05_1']
#     # outpath = Path("/home/localadmin_ccollins/harddrive4TB/parameterstudy/hesma_lc")
#     # timemin = 5
#     # timemax = 70
#     # for modelname in modelnames:
#     #     modelpath = pathtomodel / modelname
#     #     make_hesma_bol_lightcurve(modelpath, outpath, timemin, timemax)
#
#     # pathtofiles = Path("/home/localadmin_ccollins/harddrive4TB/parameterstudy/declinerate")
#     # read_hesma_peakmag_dm15_dm40(pathtofiles)
#
#
# if __name__ == '__main__':
#     main()



#!/usr/bin/env python3
import argparse
import math
# import numpy as np
from astropy import units as u
import artistools as at
import artistools.nltepops
from math import exp


def forward_doubledecay(iso1fract0, iso2fract0, iso3fract0, tlate, meanlife1_days, meanlife2_days):
    # get the abundances at a late time from the time zero abundances
    # e.g. Ni56 -> Co56 -> Fe56 decay
    # meanlife1 is the mean lifetime of the parent (e.g. Ni56) and meanlife2 is the mean life of the daughter nucleus (e.g. Co56)
    assert(tlate > 0)

    lamb1 = 1 / meanlife1_days
    lamb2 = 1 / meanlife2_days

    iso1fraclate = iso1fract0 * exp(-lamb1 * tlate)  # larger abundance before decays

    iso2fraclate = (
        iso2fract0 * exp(-lamb2 * tlate) +
        iso1fract0 * lamb1 / (lamb1 - lamb2) * (exp(-lamb2 * tlate) - exp(-lamb1 * tlate)))

    iso3fromdecay = (
        (iso1fract0 + iso2fract0) * (lamb1 - lamb2) -
        iso2fract0 * lamb1 * exp(-lamb2 * tlate) +
        iso2fract0 * lamb2 * exp(-lamb2 * tlate) -
        iso1fract0 * lamb1 * exp(-lamb2 * tlate) +
        iso1fract0 * lamb2 * exp(-lamb1 * tlate)) / (lamb1 - lamb2)

    iso3fraclate = iso3fract0 + iso3fromdecay

    return iso1fraclate, iso2fraclate, iso3fraclate


def addargs(parser):
    parser.add_argument('-modelpath', default='.',
                        help='Path to ARTIS folder')

    parser.add_argument('-timedays', '-t', default=330, type=float,
                        help='Time in days')


def main_analytical(args=None, argsraw=None, **kwargs):
    """Use the model initial conditions to calculate the deposition rates"""

    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Plot deposition rate of a model at time t (days).')
        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)
    dfmodel, t_model_init, _ = at.inputmodel.get_modeldata(args.modelpath)

    t_init = t_model_init * u.day

    meanlife_ni56 = 8.8 * u.day
    meanlife_co56 = 113.7 * u.day
    # define TCOBALT (113.7*DAY)     // cobalt-56
    # T48CR = 1.29602 * u.day
    # T48V = 23.0442 * u.day
    # define T52FE   (0.497429*DAY)
    # define T52MN   (0.0211395*DAY)

    t_now = args.timedays * u.day
    print(f't_now = {t_now.to("d")}')
    print('The following assumes that all 56Ni has decayed to 56Co and all energy comes from emitted positrons')

    # adata = at.atomic.get_levels(args.modelpath, get_photoionisations=True)
    # timestep = at.get_timestep_of_timedays(args.modelpath, args.timedays)
    # dfnltepops = at.nltepops.read_files(
    #     args.modelpath, timestep=timestep).query('Z == 26')

    # phixs = adata.query('Z==26 & ion_stage==1', inplace=False).iloc[0].levels.iloc[0].phixstable[0][1] * 1e-18

    global_posdep = 0.
    for i, row in dfmodel.iterrows():
        v_inner = row['velocity_inner'] * u.km / u.s
        v_outer = row['velocity_outer'] * u.km / u.s

        volume_init = ((4 * math.pi / 3) * ((v_outer * t_init) ** 3 - (v_inner * t_init) ** 3)).to('cm3')

        volume_now = ((4 * math.pi / 3) * ((v_outer * t_now) ** 3 - (v_inner * t_now) ** 3)).to('cm3')

        # volume_now2 = (volume_init * (t_now / t_init) ** 3).to('cm3')

        rho_init = (10 ** row['logrho']) * u.g / u.cm ** 3
        mni56_init = row['X_Ni56'] * (volume_init * rho_init).to('solMass')
        mco56_init = row['X_Co56'] * (volume_init * rho_init).to('solMass')
        mfe56_init = 0
        # mco56_now = mco56_init * math.exp(- (t_now - t_init) / meanlife_co56)
        mni56_now, mco56_now, mfe56_now = forward_doubledecay(
            mni56_init, mco56_init, mfe56_init, t_now - t_init, meanlife_ni56, meanlife_co56)

        co56_positron_dep = (0.19 * 0.610 * u.MeV * (mco56_now / (55.9398393 * u.u)) / meanlife_co56).to('erg/s')
        v48_positron_dep = 0

        global_posdep += co56_positron_dep
        power_now = co56_positron_dep + v48_positron_dep

        epsilon = power_now / volume_now
        print(f'zone {i:3d}, velocity = [{v_inner:8.2f}, {v_outer:8.2f}], epsilon = {epsilon:.3e}')
        # print(f'  epsilon = {epsilon.to("eV/(cm3s)"):.2f}')

        # dfnltepops_cell = dfnltepops.query('modelgridindex == @i', inplace=False)
        # if not dfnltepops_cell.empty:
        #     nnlevel = dfnltepops_cell.query('level == 0', inplace=False).iloc[0]['n_NLTE']
        #     width = ((v_outer - v_inner) * t_now).to('cm').value
        #     tau = width * phixs * nnlevel
        #     print(f'width: {width:.3e} cm, phixs: {phixs:.3e} cm^2, nnlevel: {nnlevel:.3e} cm^-3, tau: {tau:.3e}')
    print(f'Global posdep: {global_posdep.to("solLum"):.3e}')


def main(args=None, argsraw=None, **kwargs):
    main_analytical(args=args, argsraw=argsraw, **kwargs)
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Plot deposition rate of a model at time t (days).')
        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

        # TODO: plot deposition.out file!


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import math
# import os.path

import numpy as np
import pandas as pd
from pathlib import Path

import artistools as at


def addargs(parser):
    parser.add_argument('-outputpath', '-o',
                        default='.',
                        help='Path for output files')


def main(args=None, argsraw=None, **kwargs):
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Create solar r-process pattern in ARTIS format.')

        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

    dfsolarabund = pd.read_csv(at.config['path_datadir'] / 'solar_r_abundance_pattern.txt',
                               delim_whitespace=True, comment='#')

    dfsolarabund['radioactive'] = True

    # print(dfsolarabund)

    dfbetaminus = pd.read_csv(at.config['path_datadir'] / 'betaminusdecays.txt',
                              delim_whitespace=True, comment='#',
                              names=['A', 'Z', 'Q[MeV]', 'Egamma[MeV]', 'Eelec[MeV]',
                                     'Eneutrino[MeV]', 'tau[s]'])

    def undecayed_z(row):
        dfmasschain = dfbetaminus.query('A == @row.A', inplace=False)
        if not dfmasschain.empty:
            return int(dfmasschain.Z.min())  # decay to top of chain
        else:
            return int(row.Z)

    dfsolarabund_undecayed = dfsolarabund.copy()
    dfsolarabund_undecayed['Z'] = dfsolarabund_undecayed.apply(undecayed_z, axis=1)

    # Andreas uses 90% Fe and the rest solar
    dfsolarabund_undecayed = dfsolarabund_undecayed.append(
        {'Z': 26, 'A': 56, 'numberfrac': 0.005, 'radioactive': False}, ignore_index=True)
    dfsolarabund_undecayed = dfsolarabund_undecayed.append(
        {'Z': 27, 'A': 59, 'numberfrac': 0.005, 'radioactive': False}, ignore_index=True)
    dfsolarabund_undecayed = dfsolarabund_undecayed.append(
        {'Z': 28, 'A': 58, 'numberfrac': 0.005, 'radioactive': False}, ignore_index=True)

    normfactor = dfsolarabund_undecayed.numberfrac.sum()  # convert number fractions in solar to fractions of r-process
    dfsolarabund_undecayed.eval('numberfrac = numberfrac / @normfactor', inplace=True)

    dfsolarabund_undecayed.eval('massfrac = numberfrac * A', inplace=True)
    massfracnormfactor = dfsolarabund_undecayed.massfrac.sum()
    dfsolarabund_undecayed.eval('massfrac = massfrac / @massfracnormfactor', inplace=True)

    # print(dfsolarabund_undecayed)

    t_model_init_days = 0.000231481
    t_model_init_seconds = t_model_init_days * 24 * 60 * 60

    wollager_profilename = 'wollager_ejectaprofile_10bins.txt'
    if Path(wollager_profilename).exists():
        t_model_init_days_in = float(Path(wollager_profilename).open('rt').readline().strip().removesuffix(' day'))
        dfdensities = pd.read_csv(wollager_profilename, delim_whitespace=True, skiprows=1,
                                  names=['cellid', 'velocity_outer', 'rho'])
        dfdensities['cellid'] = dfdensities['cellid'].astype(int)
        dfdensities['velocity_inner'] = np.concatenate(([0.], dfdensities['velocity_outer'].values[:-1]))

        t_model_init_seconds_in = t_model_init_days_in * 24 * 60 * 60
        dfdensities.eval('cellmass_grams = rho * 4. / 3. * @math.pi * (velocity_outer ** 3 - velocity_inner ** 3)'
                         '* (1e5 * @t_model_init_seconds_in) ** 3', inplace=True)

        # now replace the density at the input time with the density at required time

        dfdensities.eval('rho = cellmass_grams / ('
                         '4. / 3. * @math.pi * (velocity_outer ** 3 - velocity_inner ** 3)'
                         ' * (1e5 * @t_model_init_seconds) ** 3)', inplace=True)
    else:
        dfdensities = pd.DataFrame(dict(rho=10 ** -3, velocity_outer=6.e4), index=[0])

    # print(dfdensities)
    cellcount = len(dfdensities)
    # write abundances.txt

    dictelemabund = {}
    for atomic_number in range(1, dfsolarabund_undecayed.Z.max() + 1):
        dictelemabund[f'X_{at.get_elsymbol(atomic_number)}'] = (
            dfsolarabund_undecayed.query('Z == @atomic_number', inplace=False).massfrac.sum())

    dfelabundances = pd.DataFrame([dict(inputcellid=mgi + 1, **dictelemabund) for mgi in range(cellcount)])
    # print(dfelabundances)
    at.inputmodel.save_initialabundances(dfelabundances=dfelabundances, abundancefilename=args.outputpath)

    # write model.txt

    rowdict = {
        # 'inputcellid': 1,
        # 'velocity_outer': 6.e4,
        # 'logrho': -3.,
        'X_Fegroup': 1.,
        'X_Ni56': 0.,
        'X_Co56': 0.,
        'X_Fe52': 0.,
        'X_Cr48': 0.,
        'X_Ni57': 0.,
        'X_Co57': 0.,
    }

    for _, row in dfsolarabund_undecayed.query('radioactive == True').iterrows():
        rowdict[f'X_{at.elsymbols[int(row.Z)]}{int(row.A)}'] = row.massfrac

    modeldata = []
    for mgi, densityrow in dfdensities.iterrows():
        # print(mgi, densityrow)
        modeldata.append(dict(inputcellid=mgi + 1, velocity_outer=densityrow['velocity_outer'],
                         logrho=math.log10(densityrow['rho']), **rowdict))
    # print(modeldata)

    dfmodel = pd.DataFrame(modeldata)
    # print(dfmodel)
    at.inputmodel.save_modeldata(dfmodel=dfmodel, t_model_init_days=t_model_init_days, modelpath=Path(args.outputpath))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import argparse
import glob
import multiprocessing
import sys
from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd

import artistools as at


def read_logfiles(modelpath):
    modeldata, _, _ = at.inputmodel.get_modeldata(modelpath)

    mpiranklist = at.get_mpiranklist(modelpath)
    nprocs = at.get_nprocs(modelpath)

    logfilepaths = []

    for folderpath in at.get_runfolders(modelpath):
        nfilesread_thisfolder = 0
        # for mpirank in range(nprocs):
        for mpirank in mpiranklist:
            logfilename = f"output_{mpirank}-0.txt"
            logfilepath = Path(folderpath, logfilename)
            if not logfilepath.is_file():
                # logfilepath = Path(folderpath, logfilename + '.gz')
                # if not logfilepath.is_file():
                #     print(f'Warning: Could not find {logfilepath.relative_to(modelpath.parent)}')
                continue
            logfilepaths.append(logfilepath)

    # print(logfilepaths)
    return logfilepaths


def read_time_taken(logfilepaths):
    logfiledict = {}

    updategrid_dict = {}
    updatepackets_dict = {}
    writeestimators_dict = {}

    for logfilepath in logfilepaths:
        mpi_process = int(str(logfilepath).split("/")[-1].split("-")[0].split("_")[-1])
        with open(logfilepath, "r") as logfile:

            lineswithtimes = [line.split(" ") for line in logfile if "took" in line]

        # for line in lineswithtimes:
        #     print(line)

        # get times for update_grid:
        updategridlines = [line for line in lineswithtimes if line[2] == "update_grid:"]
        for line in updategridlines:
            timestep = int(line[1].strip(":"))
            process = int(line[4])
            timetaken = int(line[-2])

            if timestep not in updategrid_dict:
                updategrid_dict[timestep] = {}
            if process not in updategrid_dict[timestep]:
                updategrid_dict[timestep][process] = timetaken

        # Get times for update packets
        updatepacketlines = [line for line in lineswithtimes if line[4] == "update" and line[5] == "packets"]
        # print(updatepacketlines)
        for line in updatepacketlines:
            timestep = int(line[1].strip(":"))
            process = mpi_process
            timetaken = int(line[-2])
            # print(timestep, process, timetaken)
            # if timestep == 30:
            # print(process, timetaken)

            if timestep not in updatepackets_dict:
                updatepackets_dict[timestep] = {}
            if process not in updatepackets_dict[timestep]:
                updatepackets_dict[timestep][process] = timetaken

        writetoestimatorslines = [line for line in lineswithtimes if line[0] == "writing" and line[2] == "estimators"]
        for line in writetoestimatorslines:
            timestep = int(line[7].split("...")[0])
            process = mpi_process
            timetaken = int(line[-2])
            # print(timestep, process, timetaken)
            # # if timestep == 30:
            # # print(process, timetaken)
            #
            if timestep not in writeestimators_dict:
                writeestimators_dict[timestep] = {}
            if process not in writeestimators_dict[timestep]:
                writeestimators_dict[timestep][process] = timetaken

    # print(updatepackets_dict[30])
    logfiledict["update_grid"] = updategrid_dict
    logfiledict["update_packets"] = updatepackets_dict
    logfiledict["write_estimators"] = writeestimators_dict
    # print(logfiledict['update_packets'][30])
    return logfiledict


def make_plot(logfiledict):
    for timestep in range(55):
        plotvalues = ["update_packets", "update_grid", "write_estimators"]
        for plotvalue in plotvalues:
            # print(logfiledict[plotvalue][timestep])
            process, timetaken = zip(*logfiledict[plotvalue][timestep].items())
            # print(process, timetaken)
            plt.plot(process, timetaken, label=plotvalue)
        plt.xlabel("mpi rank")
        plt.ylabel("time (s)")
        plt.title(f"timestep {timestep}")
        plt.legend()
        plt.show()

    # print(updategriddict.items())

    # for _ in modelgridindex[plotvalue][timestep]:
    #     print(_)


def addargs(parser):
    parser.add_argument(
        "-modelpath",
        default=[],
        nargs="*",
        action=at.AppendPath,
        help="Path to ARTIS model folders with model.txt and abundances.txt",
    )


def main(args=None, argsraw=None, **kwargs):
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter, description="Make 1D model from cone in 3D model."
        )
        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

    if not args.modelpath:
        args.modelpath = [Path(".")]

    # make_plot(logfiledict)


if __name__ == "__main__":
    multiprocessing.freeze_support()
    main()



#!/usr/bin/env python

# from rd_cmfgen import rd_nuc_decay_data
from math import exp

import numpy as np
from rd_cmfgen import rd_sn_hydro_data

# import math

msun = 1.989e33

model = "DDC25"
snapshot = "SN_HYDRO_DATA_1.300d"
# snapshot = 'SN_HYDRO_DATA_203.1d'

use_double_decay = (
    True  # should undo chains like Ni56 -> Co56 -> Fe56 instead of assuming all Fe56 and Co56 was initially Ni56
)


def undecay(a, indexofatomicnumber, indexofisotope, zparent, numnucleons):
    # e.g. parent=26, numnucleons=56 to reverse Ni56 -> Co56 decay
    daughterisofracin = a["isofrac"][:, indexofisotope[(zparent - 1, numnucleons)]]
    granddaughterisofracin = a["isofrac"][:, indexofisotope[(zparent - 2, numnucleons)]]

    a["specfrac"][:, indexofatomicnumber[zparent]] += daughterisofracin + granddaughterisofracin
    a["specfrac"][:, indexofatomicnumber[zparent - 1]] -= daughterisofracin
    a["specfrac"][:, indexofatomicnumber[zparent - 2]] -= granddaughterisofracin

    a["isofrac"][:, indexofisotope[(zparent, numnucleons)]] += daughterisofracin + granddaughterisofracin
    a["isofrac"][:, indexofisotope[(zparent - 1, numnucleons)]] -= daughterisofracin
    a["isofrac"][:, indexofisotope[(zparent - 2, numnucleons)]] -= granddaughterisofracin


def reverse_doubledecay(
    a, indexofatomicnumber, indexofisotope, zparent, numnucleons, tlate, meanlife1_days, meanlife2_days
):
    # get the abundances at time zero from the late time abundances
    # e.g. zparent=26, numnucleons=56 to reverse Ni56 -> Co56 -> Fe56 decay
    # meanlife1 is the mean lifetime of the parent (e.g. Ni56) and meanlife2 is the mean life of the daughter nucleus (e.g. Co56)
    assert tlate > 0
    iso1fraclate = a["isofrac"][:, indexofisotope[(zparent, numnucleons)]]
    iso2fraclate = a["isofrac"][:, indexofisotope[(zparent - 1, numnucleons)]]
    iso3fraclate = a["isofrac"][:, indexofisotope[(zparent - 2, numnucleons)]]

    lamb1 = 1 / meanlife1_days
    lamb2 = 1 / meanlife2_days

    iso1fract0 = np.zeros_like(iso1fraclate)
    iso2fract0 = np.zeros_like(iso1fract0)
    iso3fromdecay = np.zeros_like(iso1fract0)
    iso3fract0 = np.zeros_like(iso1fract0)
    for s in range(a["nd"]):
        iso1fract0[s] = iso1fraclate[s] * exp(lamb1 * tlate)  # larger abundance before decays

        iso2fract0[s] = (
            iso2fraclate[s] - iso1fract0[s] * lamb1 / (lamb1 - lamb2) * (exp(-lamb2 * tlate) - exp(-lamb1 * tlate))
        ) * exp(lamb2 * tlate)

        # print(iso1fract0[s], iso1fraclate[s], iso2fract0[s], iso2fraclate[s], iso1fract0[s] * lamb1 / (lamb1 - lamb2) * (exp(-lamb2 * tlate) - exp(-lamb1 * tlate)))
        # assert(iso2fract0[s] > 0)

        # print(s, (iso1fract0[s] - iso1fraclate[s]), (iso2fraclate[s] + iso3fraclate[s]))
        # print(s, (iso1fract0[s] - iso1fraclate[s]) >= (iso2fraclate[s] + iso3fraclate[s]), iso2fract0[s] < 0)

        if iso2fract0[s] < 0:
            iso2fract0[s] = iso2fraclate[s] + iso3fraclate[s] - (iso1fract0[s] - iso1fraclate[s])
            iso3fract0[s] = 0.0

            if iso2fract0[s] < 0.0:
                iso1fract0[s] += iso2fract0[s]
                iso2fract0[s] = 0.0
                print(
                    "shell",
                    s,
                    " goes fully to top isotope Z={} A={} of the chain at time zero".format(zparent, numnucleons),
                )
            else:
                print(
                    "shell",
                    s,
                    " has none of the last isotope Z={} A={} of the chain at time zero".format(
                        zparent - 2, numnucleons
                    ),
                )
        else:
            iso3fromdecay[s] = (
                (iso1fract0[s] + iso2fract0[s]) * (lamb1 - lamb2)
                - iso2fract0[s] * lamb1 * exp(-lamb2 * tlate)
                + iso2fract0[s] * lamb2 * exp(-lamb2 * tlate)
                - iso1fract0[s] * lamb1 * exp(-lamb2 * tlate)
                + iso1fract0[s] * lamb2 * exp(-lamb1 * tlate)
            ) / (lamb1 - lamb2)

            iso3fract0[s] = iso3fraclate[s] - iso3fromdecay[s]

        # print(iso2fract0[s] >= 0, iso3fract0[s] >= 0)
        sumt0 = iso1fract0[s] + iso2fract0[s] + iso3fract0[s]
        sumlate = iso1fraclate[s] + iso2fraclate[s] + iso3fraclate[s]
        if abs(sumlate - sumt0) > 1e-10:
            print(s, "t0", iso1fract0[s], iso2fract0[s], iso3fract0[s], sumt0)
            print(s, "tlate", iso1fraclate[s], iso2fraclate[s], iso3fraclate[s], sumlate, sumlate - sumt0)

        assert abs(sumt0 - sumlate) < 1e-10
        assert iso1fract0[s] >= 0.0
        assert iso2fract0[s] >= 0.0
        assert iso3fract0[s] >= 0.0

    a["isofrac"][:, indexofisotope[(zparent, numnucleons)]] = iso1fract0
    a["isofrac"][:, indexofisotope[(zparent - 1, numnucleons)]] = iso2fract0
    a["isofrac"][:, indexofisotope[(zparent - 2, numnucleons)]] = iso3fract0

    a["specfrac"][:, indexofatomicnumber[zparent]] += iso1fract0 - iso1fraclate
    a["specfrac"][:, indexofatomicnumber[zparent - 1]] += iso2fract0 - iso2fraclate
    a["specfrac"][:, indexofatomicnumber[zparent - 2]] += iso3fract0 - iso3fraclate


def forward_doubledecay(
    a, indexofatomicnumber, indexofisotope, zparent, numnucleons, tlate, meanlife1_days, meanlife2_days
):
    # get the abundances at a late time from the time zero abundances
    # e.g. zdaughter=27, numnucleons=56 for Ni56 -> Co56 -> Fe56 decay
    # meanlife1 is the mean lifetime of the parent (e.g. Ni56) and meanlife2 is the mean life of the daughter nucleus (e.g. Co56)
    assert tlate > 0
    iso1fract0 = a["isofrac"][:, indexofisotope[(zparent, numnucleons)]]
    iso2fract0 = a["isofrac"][:, indexofisotope[(zparent - 1, numnucleons)]]
    iso3fract0 = a["isofrac"][:, indexofisotope[(zparent - 2, numnucleons)]]

    lamb1 = 1 / meanlife1_days
    lamb2 = 1 / meanlife2_days

    iso1fraclate = iso1fract0 * exp(-lamb1 * tlate)  # larger abundance before decays

    iso2fraclate = iso2fract0 * exp(-lamb2 * tlate) + iso1fract0 * lamb1 / (lamb1 - lamb2) * (
        exp(-lamb2 * tlate) - exp(-lamb1 * tlate)
    )

    iso3fromdecay = (
        (iso1fract0 + iso2fract0) * (lamb1 - lamb2)
        - iso2fract0 * lamb1 * exp(-lamb2 * tlate)
        + iso2fract0 * lamb2 * exp(-lamb2 * tlate)
        - iso1fract0 * lamb1 * exp(-lamb2 * tlate)
        + iso1fract0 * lamb2 * exp(-lamb1 * tlate)
    ) / (lamb1 - lamb2)

    iso3fraclate = iso3fract0 + iso3fromdecay

    a["isofrac"][:, indexofisotope[(zparent, numnucleons)]] = iso1fraclate
    a["isofrac"][:, indexofisotope[(zparent - 1, numnucleons)]] = iso2fraclate
    a["isofrac"][:, indexofisotope[(zparent - 2, numnucleons)]] = iso3fraclate

    a["specfrac"][:, indexofatomicnumber[zparent]] += iso1fraclate - iso1fract0
    a["specfrac"][:, indexofatomicnumber[zparent - 1]] += iso2fraclate - iso2fract0
    a["specfrac"][:, indexofatomicnumber[zparent - 2]] += iso3fraclate - iso3fract0


def timeshift_double_decay(
    a, indexofatomicnumber, indexofisotope, zparent, numnucleons, timeold, timenew, meanlife1_days, meanlife2_days
):
    # take abundances back to time zero and then forward to the selected model time
    elfracsum_before = sum([a["specfrac"][:, indexofatomicnumber[zparent - i]] for i in range(3)])
    isofracsum_before = sum([a["isofrac"][:, indexofisotope[(zparent - i, numnucleons)]] for i in range(3)])

    reverse_doubledecay(
        a,
        indexofatomicnumber,
        indexofisotope,
        zparent=zparent,
        numnucleons=numnucleons,
        tlate=timeold,
        meanlife1_days=meanlife1_days,
        meanlife2_days=meanlife2_days,
    )

    forward_doubledecay(
        a,
        indexofatomicnumber,
        indexofisotope,
        zparent=zparent,
        numnucleons=numnucleons,
        tlate=timenew,
        meanlife1_days=meanlife1_days,
        meanlife2_days=meanlife2_days,
    )

    elfracsum_after = sum([a["specfrac"][:, indexofatomicnumber[zparent - i]] for i in range(3)])
    isofracsum_after = sum([a["isofrac"][:, indexofisotope[(zparent - i, numnucleons)]] for i in range(3)])
    assert np.all(abs(elfracsum_before - elfracsum_after) < 1e-10)



#!/usr/bin/env python3

import argparse
import math
import multiprocessing
import os
# import re
import sys

from astropy import constants as const
from astropy import units as u
from functools import lru_cache
from pathlib import Path
# from itertools import chain

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
# import matplotlib.patches as mpatches

import artistools as at
import artistools.spectra
import artistools.estimators
import artistools.nltepops

H = 6.6260755e-27  # Planck constant [erg s]
KB = 1.38064852e-16
EV = 1.6021772e-12  # eV to ergs [eV/erg]
ONEOVERH = 1.509188961e+26
HOVERKB = 4.799243681748932e-11
TWOOVERCLIGHTSQUARED = 2.2253001e-21
SAHACONST = 2.0706659e-16


@lru_cache(maxsize=4)
def read_files(modelpath, timestep=-1, modelgridindex=-1):
    """Read radiation field data from a list of file paths into a pandas DataFrame."""
    radfielddata = pd.DataFrame()

    mpiranklist = at.get_mpiranklist(modelpath, modelgridindex=modelgridindex)
    for folderpath in at.get_runfolders(modelpath, timestep=timestep):
        for mpirank in mpiranklist:
            radfieldfilename = f'radfield_{mpirank:04d}.out'
            radfieldfilepath = Path(folderpath, radfieldfilename)
            radfieldfilepath = at.firstexisting(
                [radfieldfilename + '.xz', radfieldfilename + '.gz', radfieldfilename], path=folderpath)

            if modelgridindex > -1:
                filesize = Path(radfieldfilepath).stat().st_size / 1024 / 1024
                print(f'Reading {Path(radfieldfilepath).relative_to(modelpath.parent)} ({filesize:.2f} MiB)')

            radfielddata_thisfile = pd.read_csv(radfieldfilepath, delim_whitespace=True)
            # radfielddata_thisfile[['modelgridindex', 'timestep']].apply(pd.to_numeric)

            if timestep >= 0:
                radfielddata_thisfile.query('timestep==@timestep', inplace=True)

            if modelgridindex >= 0:
                radfielddata_thisfile.query('modelgridindex==@modelgridindex', inplace=True)

            if not radfielddata_thisfile.empty:
                if timestep >= 0 and modelgridindex >= 0:
                    return radfielddata_thisfile
                else:
                    radfielddata = radfielddata.append(radfielddata_thisfile.copy(), ignore_index=True)

    return radfielddata


def select_bin(radfielddata, nu=None, lambda_angstroms=None, modelgridindex=None, timestep=None):
    assert nu is None or lambda_angstroms is None

    if lambda_angstroms is not None:
        nu = const.c.to('angstrom/s').value / lambda_angstroms
    else:
        lambda_angstroms = const.c.to('angstrom/s').value / nu

    dfselected = radfielddata.query(
        ('modelgridindex == @modelgridindex and ' if modelgridindex else '') +
        ('timestep == @timestep and ' if timestep else '') +
        'nu_lower <= @nu and nu_upper >= @nu and bin_num > -1')

    assert not dfselected.empty
    return dfselected.iloc[0].bin_num, dfselected.iloc[0].nu_lower, dfselected.iloc[0].nu_upper


def get_binaverage_field(axis, radfielddata, modelgridindex=None, timestep=None):
    """Get the dJ/dlambda constant average estimators of each bin."""
    # exclude the global fit parameters and detailed lines with negative "bin_num"
    bindata = radfielddata.copy().query(
        'bin_num >= 0' +
        (' & modelgridindex==@modelgridindex' if modelgridindex else '') +
        (' & timestep==@timestep' if timestep else ''))

    arr_lambda = const.c.to('angstrom/s').value / bindata['nu_upper'].values

    bindata['dlambda'] = bindata.apply(
        lambda row: const.c.to('angstrom/s').value * (1 / row['nu_lower'] - 1 / row['nu_upper']), axis=1)

    yvalues = bindata.apply(
        lambda row: row['J'] / row['dlambda'] if (
            not math.isnan(row['J'] / row['dlambda']) and row['T_R'] >= 0) else 0.0, axis=1).values

    # add the starting point
    arr_lambda = np.insert(arr_lambda, 0, const.c.to('angstrom/s').value / bindata['nu_lower'].iloc[0])
    yvalues = np.insert(yvalues, 0, 0.)

    return arr_lambda, yvalues


def j_nu_dbb(arr_nu_hz, W, T):
    """# CGS units J_nu for dilute blackbody."""
    k_b = const.k_B.to('eV/K').value
    h = const.h.to('eV s').value

    if W > 0.:
        try:
            return [W * 1.4745007e-47 * pow(nu_hz, 3) * 1.0 / (math.expm1(h * nu_hz / T / k_b)) for nu_hz in arr_nu_hz]
        except OverflowError:
            print(f"WARNING: overflow error W {W}, T {T} (Did this happen in ARTIS too?)")

    return [0. for _ in arr_nu_hz]


def get_fullspecfittedfield(radfielddata, xmin, xmax, modelgridindex=None, timestep=None):
    row = radfielddata.query(
        'bin_num == -1' +
        (' & modelgridindex==@modelgridindex' if modelgridindex else '') +
        (' & timestep==@timestep' if timestep else '')).copy().iloc[0]
    nu_lower = const.c.to('angstrom/s').value / xmin
    nu_upper = const.c.to('angstrom/s').value / xmax
    arr_nu_hz = np.linspace(nu_lower, nu_upper, num=500)
    arr_j_nu = j_nu_dbb(arr_nu_hz, row['W'], row['T_R'])

    arr_lambda = const.c.to('angstrom/s').value / arr_nu_hz
    arr_j_lambda = arr_j_nu * arr_nu_hz / arr_lambda

    return arr_lambda, arr_j_lambda


def get_fitted_field(radfielddata, modelgridindex=None, timestep=None, print_bins=False, lambdamin=None, lambdamax=None):
    """Return the fitted dilute blackbody made up of all bins."""
    arr_lambda = []
    j_lambda_fitted = []

    radfielddata_subset = radfielddata.copy().query(
        'bin_num >= 0' +
        (' & modelgridindex==@modelgridindex' if modelgridindex else '') +
        (' & timestep==@timestep' if timestep else ''))

    if lambdamin is not None:
        nu_min = const.c.to('angstrom/s').value / lambdamax

    if lambdamax is not None:
        nu_max = const.c.to('angstrom/s').value / lambdamin

    for _, row in radfielddata_subset.iterrows():
        nu_lower = row['nu_lower']
        nu_upper = row['nu_upper']

        if lambdamax is not None:
            if nu_upper > nu_max:
                continue
            nu_lower = max(nu_lower, nu_min)
        if lambdamin is not None:
            if nu_lower < nu_min:
                continue
            nu_upper = min(nu_upper, nu_max)

        if row['W'] >= 0:
            arr_nu_hz_bin = np.linspace(nu_lower, nu_upper, num=200)
            arr_j_nu = j_nu_dbb(arr_nu_hz_bin, row['W'], row['T_R'])

            arr_lambda_bin = const.c.to('angstrom/s').value / arr_nu_hz_bin
            arr_j_lambda_bin = arr_j_nu * arr_nu_hz_bin / arr_lambda_bin

            arr_lambda += list(arr_lambda_bin)
            j_lambda_fitted += list(arr_j_lambda_bin)
        else:
            arr_nu_hz_bin = [nu_lower, nu_upper]
            arr_j_lambda_bin = [0., 0.]

            arr_lambda += [const.c.to('angstrom/s').value / nu for nu in arr_nu_hz_bin]
            j_lambda_fitted += arr_j_lambda_bin

        lambda_lower = const.c.to('angstrom/s').value / row['nu_upper']
        lambda_upper = const.c.to('angstrom/s').value / row['nu_lower']
        if print_bins and (lambdamax is None or lambda_lower < lambdamax) and (lambdamin is None or lambda_upper > lambdamin):
            print(f"Bin lambda_lower {lambda_lower:.1f} W {row['W']:.1e} "
                  f"contribs {row['ncontrib']} J_nu_avg {row['J_nu_avg']:.1e}")

    return arr_lambda, j_lambda_fitted


def plot_line_estimators(axis, radfielddata, xmin, xmax, modelgridindex=None, timestep=None, **plotkwargs):
    """Plot the Jblue_lu values from the detailed line estimators on a spectrum."""
    ymax = -1

    radfielddataselected = radfielddata.query(
        'bin_num < -1' +
        (' & modelgridindex==@modelgridindex' if modelgridindex else '') +
        (' & timestep==@timestep' if timestep else ''))[['nu_upper', 'J_nu_avg']]

    const_c = const.c.to('angstrom/s').value
    radfielddataselected.eval('lambda_angstroms = @const_c / nu_upper', inplace=True)
    radfielddataselected.eval('Jb_lambda = J_nu_avg * (nu_upper ** 2) / @const_c', inplace=True)

    ymax = radfielddataselected['Jb_lambda'].max()

    if not radfielddataselected.empty:
        axis.scatter(radfielddataselected['lambda_angstroms'], radfielddataselected['Jb_lambda'],
                     label='Line estimators', s=0.2, **plotkwargs)

    return ymax


def plot_specout(axis, specfilename, timestep, peak_value=None, scale_factor=None, **plotkwargs):
    """Plot the ARTIS spectrum."""
    print(f"Plotting {specfilename}")

    if os.path.isdir(specfilename):
        modelpath = specfilename
    elif os.path.isfile(specfilename):
        modelpath = Path(specfilename).parent

    dfspectrum = at.spectra.get_spectrum(modelpath, timestep)
    label = 'Emergent spectrum'
    if scale_factor is not None:
        label += ' (scaled)'
        dfspectrum['f_lambda'] = dfspectrum['f_lambda'] * scale_factor

    if peak_value is not None:
        label += ' (normalised)'
        dfspectrum['f_lambda'] = dfspectrum['f_lambda'] / dfspectrum['f_lambda'].max() * peak_value

    dfspectrum.plot(x='lambda_angstroms', y='f_lambda', ax=axis, label=label, **plotkwargs)


@lru_cache(maxsize=128)
def evaluate_phixs(modelpath, atomic_number, lower_ion_stage, lowerlevelindex, nu_threshold, arr_nu_hz):
    adata = at.atomic.get_levels(modelpath, get_photoionisations=True)
    lower_ion_data = adata.query('Z == @atomic_number and ion_stage == @lower_ion_stage').iloc[0]
    lowerlevel = lower_ion_data.levels.iloc[lowerlevelindex]

    from scipy.interpolate import interp1d
    phixstable = lowerlevel.phixstable
    interp_sigma_bf = interp1d(
        phixstable[:, 0] * nu_threshold, phixstable[:, 1], kind='linear', bounds_error=True,
        fill_value=0., assume_sorted=True)

    def sigma_bf(nu):
        nu_factor = nu / nu_threshold
        if nu_factor < phixstable[0, 0]:
            return 0.
        elif nu_factor > phixstable[-1, 0]:
            # return 0.
            return phixstable[-1, 1] * math.pow(phixstable[-1, 0] / nu_factor, 3)

        # return interp(nu_factor, phixstable[:, 0], phixstable[:, 1], left=0.)
        return interp_sigma_bf(nu)

    return np.array([sigma_bf(nu) for nu in arr_nu_hz])


def get_kappa_bf_ion(
        atomic_number, lower_ion_stage, modelgridindex, timestep, modelpath, arr_nu_hz, max_levels):
    adata = at.atomic.get_levels(modelpath, get_photoionisations=True)
    estimators = at.estimators.read_estimators(modelpath, timestep=timestep, modelgridindex=modelgridindex)
    T_e = estimators[(timestep, modelgridindex)]['Te']

    ion_data = adata.query('Z == @atomic_number and ion_stage == @lower_ion_stage').iloc[0]
    upper_ion_data = adata.query('Z == @atomic_number and ion_stage == (@lower_ion_stage + 1)').iloc[0]

    lowerionpopdensity = estimators[(timestep, modelgridindex)]['populations'][(atomic_number, lower_ion_stage)]

    ion_popfactor_sum = 0
    for _, level in ion_data.levels[:max_levels].iterrows():
        ion_popfactor_sum += level.g * math.exp(-level.energy_ev * EV / KB / T_e)

    array_kappa_bf_nu_ion = np.zeros_like(arr_nu_hz)
    for levelnum, lowerlevel in ion_data.levels[:max_levels].iterrows():
        levelpopfrac = lowerlevel.g * math.exp(-lowerlevel.energy_ev * EV / KB / T_e) / ion_popfactor_sum

        for upperlevelnum, phixsfrac in lowerlevel.phixstargetlist:
            upperlevel = upper_ion_data.levels.iloc[upperlevelnum]

            nu_threshold = ONEOVERH * (ion_data.ion_pot - lowerlevel.energy_ev + upperlevel.energy_ev) * EV

            arr_sigma_bf = evaluate_phixs(modelpath, atomic_number, lower_ion_stage, levelnum,
                                          nu_threshold, tuple(arr_nu_hz)) * phixsfrac

            array_kappa_bf_nu_ion += arr_sigma_bf * levelpopfrac * lowerionpopdensity

    return array_kappa_bf_nu_ion


def get_recombination_emission(
        atomic_number, upper_ion_stage, arr_nu_hz, modelgridindex, timestep, modelpath, max_levels, use_lte_pops=False):
    adata = at.atomic.get_levels(modelpath, get_photoionisations=True)

    lower_ion_stage = upper_ion_stage - 1
    upperionstr = at.get_ionstring(atomic_number, upper_ion_stage)
    lowerionstr = at.get_ionstring(atomic_number, lower_ion_stage)
    upper_ion_data = adata.query('Z == @atomic_number and ion_stage == @upper_ion_stage').iloc[0]
    lower_ion_data = adata.query('Z == @atomic_number and ion_stage == @lower_ion_stage').iloc[0]

    estimators = at.estimators.read_estimators(modelpath, timestep=timestep, modelgridindex=modelgridindex)

    upperionpopdensity = estimators[(timestep, modelgridindex)]['populations'][(atomic_number, upper_ion_stage)]
    T_e = estimators[(timestep, modelgridindex)]['Te']
    nne = estimators[(timestep, modelgridindex)]['nne']

    upperionpopdensity = estimators[(timestep, modelgridindex)]['populations'][(atomic_number, upper_ion_stage)]
    print(f'Recombination from {upperionstr} -> {lowerionstr} ({upperionstr} pop = {upperionpopdensity:.1e}/cm3)')

    if use_lte_pops:
        upper_level_popfactor_sum = 0
        for upperlevelnum, upperlevel in lower_ion_data.levels[:200].iterrows():
            upper_level_popfactor_sum += upperlevel.g * math.exp(-upperlevel.energy_ev * EV / KB / T_e)
    else:
        dfnltepops = at.nltepops.read_files(modelpath, modelgridindex=modelgridindex, timestep=timestep)
        dfnltepops_upperion = dfnltepops.query('Z==@atomic_number & ion_stage==@upper_ion_stage')
        upperion_nltepops = {x.level: x['n_NLTE'] for _, x in dfnltepops_upperion.iterrows()}

    arr_j_nu_lowerlevel = {}
    arr_alpha_dnu = np.zeros_like(arr_nu_hz)
    alpha_ion2 = 0.
    nnionfrac = 0.
    for levelnum, lowerlevel in lower_ion_data.levels[:max_levels].iterrows():
        for upperlevelnum, phixsfrac in lowerlevel.phixstargetlist:
            upperlevel = upper_ion_data.levels.iloc[upperlevelnum]

            if use_lte_pops:
                levelpopfrac = (
                    upperlevel.g * math.exp(-upperlevel.energy_ev * EV / KB / T_e) / upper_level_popfactor_sum)
            elif len(upperion_nltepops) == 1:  # top ion has only one level
                levelpopfrac = upperion_nltepops[0] / upperionpopdensity
            else:
                levelpopfrac = upperion_nltepops[upperlevelnum] / upperionpopdensity

            nu_threshold = ONEOVERH * (lower_ion_data.ion_pot - lowerlevel.energy_ev + upperlevel.energy_ev) * EV

            arr_sigma_bf = evaluate_phixs(modelpath, atomic_number, lower_ion_stage, levelnum,
                                          nu_threshold, tuple(arr_nu_hz)) * phixsfrac

            sfac = (SAHACONST * lowerlevel.g / upperlevel.g *
                    math.pow(T_e, -1.5) * math.exp(HOVERKB * nu_threshold / T_e))

            arr_alpha_level_dnu = 4. * math.pi * sfac * (
                TWOOVERCLIGHTSQUARED * arr_sigma_bf *
                np.power(arr_nu_hz, 2) * np.exp(-HOVERKB * arr_nu_hz / T_e)) * levelpopfrac

            arr_j_nu_lowerlevel[(upperlevelnum, levelnum)] = arr_alpha_level_dnu / 4 / math.pi * H * arr_nu_hz * upperionpopdensity * nne

            arr_alpha_dnu += arr_alpha_level_dnu

            nnionfrac += levelpopfrac

            # arr_nu_hz2 = nu_threshold * lowerlevel.phixstable[:, 0]
            arr_nu_hz2 = nu_threshold * np.linspace(1.0, 1.0 + 0.03 * (100 + 1), num=3 * 100 + 1, endpoint=False)
            arr_sigma_bf2 = evaluate_phixs(modelpath, atomic_number, lower_ion_stage, levelnum,
                                           nu_threshold, tuple(arr_nu_hz2)) * phixsfrac
            arr_alpha_level_dnu2 = 4. * math.pi * sfac * (
                TWOOVERCLIGHTSQUARED * arr_sigma_bf2 *
                np.power(arr_nu_hz2, 2) * np.exp(-HOVERKB * arr_nu_hz2 / T_e)) * levelpopfrac
            alpha_level2 = np.abs(np.trapz(arr_alpha_level_dnu2, x=arr_nu_hz2))
            alpha_ion2 += alpha_level2

            # alpha_level = np.abs(np.trapz(arr_alpha_level_dnu, x=arr_nu_hz))
            lambda_threshold = const.c.to('angstrom/s').value / nu_threshold
            print(f' {upperionstr} level {upperlevelnum} -> {lowerionstr} level {levelnum}'
                  f' threshold {lambda_threshold:7.1f} Å'
                  f' Alpha_R_contrib {alpha_level2:.2e} {lowerlevel.levelname}'
                  f' upperlevelpop {upperion_nltepops[upperlevelnum]:.2e}')

    alpha_ion = np.abs(np.trapz(arr_alpha_dnu, x=arr_nu_hz))
    print(f'  {upperionstr} Alpha_R = {alpha_ion:.2e}   Alpha_R*nne = {nne*alpha_ion:.2e}')
    print(f'  {upperionstr} Alpha_R2 = {alpha_ion2:.2e} Alpha_R2*nne = {nne*alpha_ion2:.2e}')
    # c_cgs = const.c.to('cm/s').value
    # vmax = at.inputmodel.get_modeldata(modelpath)[0]['velocity_outer'].iloc[-1] * 1e5
    # t_seconds = at.get_timestep_times_float(modelpath, loc='start')[timestep] * u.day.to('s')

    # mean_free_path = vmax * t_seconds

    arr_j_nu = arr_alpha_dnu / 4 / math.pi * H * arr_nu_hz * upperionpopdensity * nne
    return arr_j_nu, arr_j_nu_lowerlevel


def get_ion_gamma_dnu(modelpath, modelgridindex, timestep, atomic_number, ion_stage, arr_nu_hz, J_nu_arr, max_levels):
    """Calculate the contribution to the photoionisation rate coefficient per J_nu at each frequency nu for an ion."""
    estimators = at.estimators.read_estimators(modelpath, timestep=timestep, modelgridindex=modelgridindex)

    T_e = estimators[(timestep, modelgridindex)]['Te']
    T_R = estimators[(timestep, modelgridindex)]['TR']

    adata = at.atomic.get_levels(modelpath, get_photoionisations=True)
    ion_data = adata.query('Z == @atomic_number and ion_stage == @ion_stage').iloc[0]
    upper_ion_data = adata.query('Z == @atomic_number and ion_stage == (@ion_stage + 1)').iloc[0]
    ionstr = at.get_ionstring(atomic_number, ion_stage)

    ion_popfactor_sum = 0
    for _, level in ion_data.levels[:max_levels].iterrows():
        ion_popfactor_sum += level.g * math.exp(-level.energy_ev * EV / KB / T_e)

    arr_gamma_dnu = np.zeros_like(arr_nu_hz)
    for levelnum, level in ion_data.levels[:max_levels].iterrows():
        levelpopfrac = level.g * math.exp(-level.energy_ev * EV / KB / T_e) / ion_popfactor_sum

        for upperlevelnum, phixsfrac in level.phixstargetlist:
            upperlevel = upper_ion_data.levels.iloc[upperlevelnum]
            nu_threshold = ONEOVERH * (ion_data.ion_pot - level.energy_ev + upperlevel.energy_ev) * EV

            arr_sigma_bf = evaluate_phixs(modelpath, atomic_number, ion_stage,
                                          levelnum, nu_threshold, tuple(arr_nu_hz)) * phixsfrac

            arr_corrfactors = 1 - np.exp(-HOVERKB * arr_nu_hz / T_R)
            assert(min(arr_corrfactors) > 0.50)
            assert(max(arr_corrfactors) <= 1.)

            arr_gamma_level_dnu = (
                4 * math.pi * ONEOVERH * arr_sigma_bf / arr_nu_hz * J_nu_arr *
                arr_corrfactors * levelpopfrac)

            arr_gamma_dnu += arr_gamma_level_dnu

            gamma_r_level = np.abs(np.trapz(arr_gamma_level_dnu, x=arr_nu_hz))
            lambda_threshold = const.c.to('angstrom/s').value / nu_threshold

            print(f'  level {levelnum} pop_frac {levelpopfrac:.2f} upperlevel {upperlevelnum}'
                  f' threshold {lambda_threshold:.1f} Å '
                  f' gamma_R_level({ionstr}) {gamma_r_level:.2e} {level.levelname}')

    return arr_gamma_dnu


def calculate_photoionrates(axes, modelpath, radfielddata, modelgridindex, timestep, xmin, xmax, ymax, args):
    axes[0].set_ylabel(r'$\sigma$ [cm$^2$]')

    # recomblowerionlist = ((26, 1), (26, 2), (26, 3), (26, 4), (27, 2), (27, 3), (28, 2), (28, 3), (28, 4))
    # photoionlist = ((26, 2), (28, 2))

    recomblowerionlist = ((26, 3),)
    photoionlist = ((26, 2),)
    kappalowerionlist = ((26, 2), (26, 3),)
    adata = at.atomic.get_levels(modelpath, get_photoionisations=True)

    fieldlist = []

    arr_lambda_fitted, j_lambda_fitted = get_fitted_field(
        radfielddata, modelgridindex=modelgridindex, timestep=timestep,
        print_bins=True, lambdamin=xmin, lambdamax=xmax)

    arr_lambda_fitted, j_lambda_fitted = zip(*[
        pt for pt in zip(arr_lambda_fitted, j_lambda_fitted) if xmin <= pt[0] <= xmax])

    estimators = at.estimators.read_estimators(modelpath, timestep=timestep, modelgridindex=modelgridindex)

    max_levels = 20

    T_e = estimators[(timestep, modelgridindex)]['Te']
    nne = estimators[(timestep, modelgridindex)]['nne']
    print(f'T_e {T_e:.1f} K, nne {nne:.1e} /cm3')

    arraylambda_angstrom_recomb = np.linspace(xmin, xmax, num=10000)
    arr_nu_hz_recomb = const.c.to('angstrom/s').value / np.array(arraylambda_angstrom_recomb)

    # calculate bound-free opacity
    array_kappa_bf_nu = np.zeros_like(arr_nu_hz_recomb)
    for atomic_number, lower_ion_stage in kappalowerionlist:
        array_kappa_bf_nu += get_kappa_bf_ion(
            atomic_number, lower_ion_stage, modelgridindex, timestep, modelpath, arr_nu_hz_recomb, max_levels)

    # calculate recombination emission
    J_lambda_recomb_total = np.zeros_like(arraylambda_angstrom_recomb)

    lw = 1.
    for atomic_number, lower_ion_stage in recomblowerionlist:
        # lw -= 0.1
        upperionstr = at.get_ionstring(atomic_number, lower_ion_stage + 1)

        j_emiss_nu_recomb, arr_j_nu_lowerleveldict = get_recombination_emission(
            atomic_number, lower_ion_stage + 1, arr_nu_hz_recomb,
            modelgridindex, timestep, modelpath, max_levels)

        for (upperlevelnum, lowerlevelnum), arr_j_emiss_nu_lowerlevel in arr_j_nu_lowerleveldict.items():
            J_nu_recomb = np.array([j_emiss_nu / kappa_bf if j_emiss_nu > 0 else 0.
                                    for j_emiss_nu, kappa_bf in zip(arr_j_emiss_nu_lowerlevel, array_kappa_bf_nu)])

            J_contrib = np.abs(np.trapz(J_nu_recomb, x=arr_nu_hz_recomb))

            J_lambda_recomb_level = J_nu_recomb * arr_nu_hz_recomb / arraylambda_angstrom_recomb
            fieldlabel = f'{upperionstr} level {upperlevelnum} -> {at.roman_numerals[lower_ion_stage]} level {lowerlevelnum}'
            axes[2].plot(arraylambda_angstrom_recomb, J_lambda_recomb_level, label=fieldlabel, lw=lw)
            fieldlist += [(arraylambda_angstrom_recomb, J_lambda_recomb_level, fieldlabel)]

        # calculate intensity from emission and opacity to (assuming dJ/ds = 0)
        J_nu_recomb = np.array([j_emiss_nu / kappa_bf if j_emiss_nu > 0 else 0.
                                for j_emiss_nu, kappa_bf in zip(j_emiss_nu_recomb, array_kappa_bf_nu)])

        J_contrib = np.abs(np.trapz(J_nu_recomb, x=arr_nu_hz_recomb))

        J_lambda_recomb = J_nu_recomb * arr_nu_hz_recomb / arraylambda_angstrom_recomb

        print(f'  Contribution to J = {J_contrib:.1e} ergs/s/cm2')

        J_lambda_recomb_total += J_lambda_recomb

        # contribution of all levels of the ion
        fieldlabel = f'{upperionstr} -> {at.roman_numerals[lower_ion_stage]} recombination'
        axes[2].plot(arraylambda_angstrom_recomb, J_lambda_recomb, label=fieldlabel, lw=lw)
        fieldlist += [(arraylambda_angstrom_recomb, J_lambda_recomb, fieldlabel)]

    # fieldlabel = f'Summed recombination'
    # axes[2].plot(arraylambda_angstrom_recomb, J_lambda_recomb_total, label=fieldlabel)
    # fieldlist += [(arraylambda_angstrom_recomb, J_lambda_recomb_total, fieldlabel)]
    ymax = max(ymax, max(J_lambda_recomb_total))

    if args.frompackets:
        (contribution_list, array_jlambda_emission_total,
         arraylambda_angstrom_em) = at.spectra.get_flux_contributions_from_packets(
            modelpath, timelowerdays=-1, timeupperdays=2000, lambda_min=xmin, lambda_max=xmax,
            getemission=True, getabsorption=False, modelgridindex=modelgridindex,
            maxpacketfiles=args.maxpacketfiles,
            groupby='ion', delta_lambda=20,
            useinternalpackets=True)
        axes[2].plot(arraylambda_angstrom_em, array_jlambda_emission_total, label='Internal radfield from packets')

        fieldlist += list([(arraylambda_angstrom_em, contribrow.array_flambda_emission, contribrow.linelabel)
                          for contribrow in contribution_list])
        fieldlist += [(arraylambda_angstrom_em, array_jlambda_emission_total, 'Total emission')]

    # fieldlist += [(arr_lambda_fitted, j_lambda_fitted, 'binned field')]

    for atomic_number, ion_stage in photoionlist:
        ionstr = at.get_ionstring(atomic_number, ion_stage)
        ion_data = adata.query('Z == @atomic_number and ion_stage == @ion_stage').iloc[0]

        for levelnum, level in ion_data.levels[:max_levels].iterrows():
            nu_threshold = ONEOVERH * (ion_data.ion_pot - level.energy_ev) * EV
            arr_sigma_bf = evaluate_phixs(modelpath, atomic_number, ion_stage, levelnum,
                                          nu_threshold, tuple(arr_nu_hz_recomb))
            if levelnum < 5:
                axes[0].plot(arraylambda_angstrom_recomb, arr_sigma_bf,
                             label=r'$\sigma_{bf}$' + f'({ionstr} {level.levelname})')

        lw = 1.
        for arraylambda_angstrom, J_lambda_arr, linelabel in fieldlist:
            # lw -= 0.4
            arr_nu_hz = const.c.to('angstrom/s').value / np.array(arraylambda_angstrom)
            print(f'{ionstr} photoionisation rate coeffs using radiation field due to {linelabel}')
            J_nu_arr = np.array(J_lambda_arr) * arraylambda_angstrom / arr_nu_hz

            arr_gamma_dnu = get_ion_gamma_dnu(modelpath, modelgridindex, timestep,
                                              atomic_number, ion_stage, arr_nu_hz, J_nu_arr, max_levels)

            # xlist = arr_lambda_fitted
            arr_gamma_dlambda = arr_gamma_dnu * arr_nu_hz / arraylambda_angstrom

            axes[1].plot(arraylambda_angstrom, arr_gamma_dlambda, lw=lw,
                         label=r'd$\Gamma_R$(' + ionstr + ' due to ' + linelabel + r')/d$\lambda$')

            gamma_r_ion = abs(np.trapz(arr_gamma_dlambda, x=arraylambda_angstrom))
            print(f'  Gamma_R({ionstr} due to {linelabel}): {gamma_r_ion:.2e}')

    axes[0].set_yscale('log')
    # axes[1].set_yscale('log')
    # axes[2].set_yscale('log')
    return ymax


def get_binedges(radfielddata):
    return [const.c.to('angstrom/s').value / radfielddata['nu_lower'].iloc[1]] + \
               list(const.c.to('angstrom/s').value / radfielddata['nu_upper'][1:])


def plot_celltimestep(
        modelpath, timestep, outputfile, xmin, xmax, modelgridindex, args, normalised=False):
    """Plot a cell at a timestep things like the bin edges, fitted field, and emergent spectrum (from all cells)."""
    radfielddata = read_files(modelpath, timestep=timestep, modelgridindex=modelgridindex)
    if radfielddata.empty:
        print(f'No data for timestep {timestep:d} modelgridindex {modelgridindex:d}')
        return False

    modelname = at.get_model_name(modelpath)
    time_days = at.get_timestep_times_float(modelpath)[timestep]
    print(f'Plotting {modelname} timestep {timestep:d} (t={time_days:.3f}d)')
    T_R = radfielddata.query('bin_num == -1').iloc[0].T_R
    print(f'T_R = {T_R}')

    nrows = 1 if not args.photoionrates else 3
    fig, axes = plt.subplots(nrows=nrows, ncols=1, sharex=True,
                             figsize=(args.figscale * at.figwidth, args.figscale * at.figwidth * (0.25 + nrows * 0.4)),
                             tight_layout={"pad": 0.2, "w_pad": 0.0, "h_pad": 0.0})

    axis = axes if nrows == 1 else axes[-1]
    ymax = 0.

    xlist, yvalues = get_fullspecfittedfield(
        radfielddata, xmin, xmax, modelgridindex=modelgridindex, timestep=timestep)

    label = r'Dilute blackbody model '
    # label += r'(T$_{\mathrm{R}}$'
    # label += f'= {row["T_R"]} K)')
    axis.plot(xlist, yvalues, label=label, color='purple', linewidth=1.5)
    ymax = max(yvalues)

    if not args.nobandaverage:
        arr_lambda, yvalues = get_binaverage_field(
            axis, radfielddata, modelgridindex=modelgridindex, timestep=timestep)
        axis.step(arr_lambda, yvalues, where='pre', label='Band-average field', color='green', linewidth=1.5)
        ymax = max([ymax] + [point[1] for point in zip(arr_lambda, yvalues) if xmin <= point[0] <= xmax])

    arr_lambda_fitted, j_lambda_fitted = get_fitted_field(
        radfielddata, modelgridindex=modelgridindex, timestep=timestep)
    ymax = max([ymax] + [point[1] for point in zip(arr_lambda_fitted, j_lambda_fitted) if xmin <= point[0] <= xmax])

    axis.plot(arr_lambda_fitted, j_lambda_fitted, label='Radiation field model', alpha=0.8, color='blue', linewidth=1.5)

    ymax3 = plot_line_estimators(
        axis, radfielddata, xmin, xmax, modelgridindex=modelgridindex, timestep=timestep, zorder=-2, color='red')

    ymax = max(ymax, ymax3)
    if args.ymax >= 0:
        ymax = args.ymax

    try:
        specfilename = at.firstexisting(['spec.out', 'spec.out.gz'], path=modelpath)
    except FileNotFoundError:
        print('Could not find spec.out')
        args.nospec = True

    if not args.nospec:
        plotkwargs = {}
        if not normalised:
            modeldata, _, t_model_init = at.inputmodel.get_modeldata(modelpath)
            # outer velocity
            v_surface = modeldata.loc[int(radfielddata.modelgridindex.max())].velocity_outer * u.km / u.s
            r_surface = (time_days * u.day * v_surface).to('km')
            r_observer = u.megaparsec.to('km')
            scale_factor = (r_observer / r_surface) ** 2 / (2 * math.pi)
            print(f'Scaling emergent spectrum flux at 1 Mpc to specific intensity '
                  f'at surface (v={v_surface:.3e}, r={r_surface:.3e})')
            plotkwargs['scale_factor'] = scale_factor
        else:
            plotkwargs['peak_value'] = ymax

        plot_specout(axis, specfilename, timestep, zorder=-1,
                     color='black', alpha=0.6, linewidth=1.0, **plotkwargs)

    if args.showbinedges:
        binedges = get_binedges(radfielddata)
        axis.vlines(binedges, ymin=0.0, ymax=ymax, linewidth=0.5,
                    color='red', label='', zorder=-1, alpha=0.4)

    velocity = at.inputmodel.get_modeldata(modelpath)[0]['velocity_outer'][modelgridindex]

    figure_title = f'{modelname} {velocity:.0f} km/s at {time_days:.0f}d'
    # figure_title += '\ncell {modelgridindex} timestep {timestep}'

    if not args.notitle:
        axis.set_title(figure_title, fontsize=11)

    if args.photoionrates:
        ymax = calculate_photoionrates(
            axes, modelpath, radfielddata, modelgridindex=modelgridindex, timestep=timestep,
            xmin=xmin, xmax=xmax, ymax=ymax, args=args)
        axes[0].legend(loc='best', handlelength=2, frameon=False, numpoints=1, fontsize=4)
        axes[1].legend(loc='best', handlelength=2, frameon=False, numpoints=1, fontsize=4)

    # axis.annotate(figure_title,
    #               xy=(0.02, 0.96), xycoords='axes fraction',
    #               horizontalalignment='left', verticalalignment='top', fontsize=8)

    axis.set_xlabel(r'Wavelength ($\mathrm{{\AA}}$)')
    axis.set_ylabel(r'J$_\lambda$ [{}erg/s/cm$^2$/$\mathrm{{\AA}}$]')
    axis.xaxis.set_minor_locator(ticker.MultipleLocator(base=500))
    axis.set_xlim(left=xmin, right=xmax)
    axis.set_ylim(bottom=0.0, top=ymax)
    axis.yaxis.set_major_formatter(at.ExponentLabelFormatter(axis.get_ylabel(), useMathText=True))

    axis.legend(loc='best', handlelength=2, frameon=False, numpoints=1, fontsize=9)

    print(f'Saving to {outputfile}')
    fig.savefig(str(outputfile), format='pdf')
    plt.close()
    return True


def plot_bin_fitted_field_evolution(axis, radfielddata, nu_line, modelgridindex, **plotkwargs):
    bin_num, _nu_lower, _nu_upper = select_bin(radfielddata, nu=nu_line, modelgridindex=modelgridindex)
    # print(f"Selected bin_num {bin_num} to get a binned radiation field estimator")

    radfielddataselected = radfielddata.query(
        'bin_num == @bin_num and modelgridindex == @modelgridindex'
        ' and nu_lower <= @nu_line and nu_upper >= @nu_line').copy()

    radfielddataselected['Jb_nu_at_line'] = radfielddataselected.apply(
        lambda x: j_nu_dbb([nu_line], x.W, x.T_R)[0], axis=1)

    const_c = const.c.to('angstrom/s').value
    radfielddataselected.eval('Jb_lambda_at_line = Jb_nu_at_line * (@nu_line ** 2) / @const_c', inplace=True)
    lambda_angstroms = const_c / nu_line

    radfielddataselected.plot(x='timestep', y='Jb_lambda_at_line', ax=axis,
                              label=f'Fitted field from bin at {lambda_angstroms:.1f} Å', **plotkwargs)


def plot_global_fitted_field_evolution(axis, radfielddata, nu_line, modelgridindex, **plotkwargs):
    radfielddataselected = radfielddata.query('bin_num == -1 and modelgridindex == @modelgridindex').copy()

    radfielddataselected['J_nu_fullspec_at_line'] = radfielddataselected.apply(
        lambda x: j_nu_dbb([nu_line], x.W, x.T_R)[0], axis=1)

    const_c = const.c.to('angstrom/s').value
    radfielddataselected.eval('J_lambda_fullspec_at_line = J_nu_fullspec_at_line * (@nu_line ** 2) / @const_c',
                              inplace=True)
    lambda_angstroms = const_c / nu_line

    radfielddataselected.plot(x='timestep', y='J_lambda_fullspec_at_line', ax=axis,
                              label=f'Full-spec fitted field at {lambda_angstroms:.1f} Å', **plotkwargs)


def plot_line_estimator_evolution(axis, radfielddata, bin_num, modelgridindex=None,
                                  timestep_min=None, timestep_max=None, **plotkwargs):
    """Plot the Jblue_lu values over time for a detailed line estimators."""
    radfielddataselected = radfielddata.query(
        'bin_num == @bin_num' +
        (' & modelgridindex == @modelgridindex' if modelgridindex else '') +
        (' & timestep >= @timestep_min' if timestep_min else '') +
        (' & timestep <= @timestep_max' if timestep_max else ''))[['timestep', 'nu_upper', 'J_nu_avg']]

    const_c = const.c.to('angstrom/s').value
    radfielddataselected.eval('lambda_angstroms = @const_c / nu_upper', inplace=True)
    radfielddataselected.eval('Jb_lambda = J_nu_avg * (nu_upper ** 2) / @const_c', inplace=True)

    axis.plot(radfielddataselected['timestep'], radfielddataselected['Jb_lambda'],
              label=f'Jb_lu bin_num {bin_num}', **plotkwargs)


def plot_timeevolution(modelpath, outputfile, modelgridindex, args):
    """Plot a estimator evolution over time for a cell. This is not well tested and should be checked."""
    print(f'Plotting time evolution of cell {modelgridindex:d}')

    radfielddata = read_files(modelpath, modelgridindex=modelgridindex)
    radfielddataselected = radfielddata.query('modelgridindex == @modelgridindex')

    const_c = const.c.to('angstrom/s').value

    nlinesplotted = 200
    fig, axes = plt.subplots(nlinesplotted, 1, sharex=True,
                             figsize=(args.figscale * at.figwidth,
                                      args.figscale * at.figwidth * (0.25 + nlinesplotted * 0.35)),
                             tight_layout={"pad": 0.2, "w_pad": 0.0, "h_pad": 0.0})

    timestep = at.get_timestep_of_timedays(modelpath, 330)
    time_days = float(at.get_timestep_time(modelpath, timestep))

    dftopestimators = radfielddataselected.query('timestep==@timestep and bin_num < -1').copy()
    dftopestimators.eval('lambda_angstroms = @const_c / nu_upper', inplace=True)
    dftopestimators.eval('Jb_lambda = J_nu_avg * (nu_upper ** 2) / @const_c', inplace=True)
    dftopestimators.sort_values(by='Jb_lambda', ascending=False, inplace=True)
    dftopestimators = dftopestimators.iloc[0:nlinesplotted]

    print(f'Top estimators at timestep {timestep} t={time_days:.1f}')
    print(dftopestimators)

    for ax, bin_num_estimator, nu_line in zip(axes, dftopestimators.bin_num.values, dftopestimators.nu_upper.values):
        lambda_angstroms = const_c / nu_line
        print(f"Selected line estimator with bin_num {bin_num_estimator}, lambda={lambda_angstroms:.1f}")
        plot_line_estimator_evolution(ax, radfielddataselected, bin_num_estimator, modelgridindex=modelgridindex)

        plot_bin_fitted_field_evolution(ax, radfielddata, nu_line, modelgridindex=modelgridindex)

        plot_global_fitted_field_evolution(ax, radfielddata, nu_line, modelgridindex=modelgridindex)
        ax.annotate(
            r'$\lambda$='
            f'{lambda_angstroms:.1f} Å in cell {modelgridindex:d}\n',
            xy=(0.02, 0.96), xycoords='axes fraction',
            horizontalalignment='left', verticalalignment='top', fontsize=10)

        ax.set_ylabel(r'J$_\lambda$ [erg/s/cm$^2$/$\mathrm{{\AA}}$]')
        ax.legend(loc='best', handlelength=2, frameon=False, numpoints=1)

    axes[-1].set_xlabel(r'Timestep')
    # axis.xaxis.set_minor_locator(ticker.MultipleLocator(base=100))
    # axis.set_xlim(left=xmin, right=xmax)
    # axis.set_ylim(bottom=0.0, top=ymax)

    print(f'Saving to {outputfile}')
    fig.savefig(str(outputfile), format='pdf')
    plt.close()


def addargs(parser):
    """Add arguments to an argparse parser object."""
    parser.add_argument('-modelpath', default='.', type=Path,
                        help='Path to ARTIS folder')

    parser.add_argument('-listtimesteps', action='store_true',
                        help='Show the times at each timestep')

    parser.add_argument('--frompackets', action='store_true',
                        help='Read packets files for internal radiation field')

    parser.add_argument('-maxpacketfiles', type=int, default=None,
                        help='Limit the number of packet files read')

    parser.add_argument('-xaxis', '-x', default='lambda', choices=['lambda', 'timestep'],
                        help='Horizontal axis variable.')

    parser.add_argument('-timedays', '-time', '-t',
                        help='Time in days to plot')

    parser.add_argument('-timestep', '-ts', action='append',
                        help='Timestep number to plot')

    parser.add_argument('-modelgridindex', '-cell', action='append',
                        help='Modelgridindex to plot')

    parser.add_argument('-velocity', '-v', type=float, default=-1,
                        help='Specify cell by velocity')

    parser.add_argument('--nospec', action='store_true',
                        help='Don\'t plot the emergent specrum')

    parser.add_argument('--showbinedges', action='store_true',
                        help='Plot vertical lines at the bin edges')

    parser.add_argument('-xmin', type=int, default=1000,
                        help='Plot range: minimum wavelength in Angstroms')

    parser.add_argument('-xmax', type=int, default=20000,
                        help='Plot range: maximum wavelength in Angstroms')

    parser.add_argument('-ymax', type=int, default=-1,
                        help='Plot range: maximum J_nu')

    parser.add_argument('--normalised', action='store_true',
                        help='Normalise the spectra to their peak values')

    parser.add_argument('--notitle', action='store_true',
                        help='Suppress the top title from the plot')

    parser.add_argument('--nobandaverage', action='store_true',
                        help='Suppress the band-average line')

    parser.add_argument('--photoionrates', action='store_true',
                        help='Suppress the band-average line')

    parser.add_argument('-figscale', type=float, default=1.,
                        help='Scale factor for plot area. 1.0 is for single-column')

    parser.add_argument('-o', action='store', dest='outputfile', type=Path,
                        help='Filename for PDF file')


def main(args=None, argsraw=None, **kwargs):
    """Plot the radiation field estimators."""
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Plot ARTIS internal radiation field estimators.')
        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

    if args.xaxis == 'lambda':
        defaultoutputfile = Path('plotradfield_cell{modelgridindex:03d}_ts{timestep:03d}.pdf')
    else:
        defaultoutputfile = Path('plotradfield_cell{modelgridindex:03d}_evolution.pdf')

    if not args.outputfile:
        args.outputfile = defaultoutputfile
    elif args.outputfile.is_dir():
        args.outputfile = args.outputfile / defaultoutputfile

    modelpath = args.modelpath

    pdf_list = []
    modelpath_list = []
    modelgridindexlist = []

    if args.listtimesteps:
        at.showtimesteptimes(modelpath=args.modelpath)
    else:
        if args.velocity >= 0.:
            modelgridindexlist = [at.inputmodel.get_mgi_of_velocity_kms(modelpath, args.velocity)]
        else:
            if args.modelgridindex is None:
                modelgridindexlist = [0]
            else:
                modelgridindexlist = at.parse_range_list(args.modelgridindex)

        timesteplast = len(at.get_timestep_times_float(modelpath))
        if args.timedays:
            timesteplist = [at.get_timestep_of_timedays(modelpath, args.timedays)]
        elif args.timestep:
            timesteplist = at.parse_range_list(args.timestep, dictvars={'last': timesteplast})
        else:
            print("Using last timestep.")
            timesteplist = [timesteplast]

        for modelgridindex in modelgridindexlist:
            if args.xaxis == 'lambda':
                for timestep in timesteplist:
                    outputfile = str(args.outputfile).format(modelgridindex=modelgridindex, timestep=timestep)
                    make_plot = plot_celltimestep(
                        modelpath, timestep, outputfile,
                        xmin=args.xmin, xmax=args.xmax, modelgridindex=modelgridindex,
                        args=args, normalised=args.normalised)
                    if make_plot:
                        pdf_list.append(outputfile)
                        modelpath_list.append(args.modelpath)
            elif args.xaxis == 'timestep':
                outputfile = args.outputfile.format(modelgridindex=modelgridindex)
                plot_timeevolution(modelpath, outputfile, modelgridindex, args)
            else:
                print('Unknown plot type {args.plot}')
                return 1

    if len(pdf_list) > 1:
        print(pdf_list, modelpath_list)
        at.join_pdf_files(pdf_list, modelpath_list)

    return 0


if __name__ == "__main__":
    multiprocessing.freeze_support()
    sys.exit(main())



'''various functions to read/write CMFGEN input/output files:

rd_nuc_decay_data
rd_sn_hydro_data
'''

# import os
import sys

# import re
import numpy as np

# from pdb import set_trace as stop

# constants
DAY2SEC = 86400.0
MEV2ERG = 1.60217733e-6


def rd_nuc_decay_data(file, quiet=False):
    """read in NUC_DECAY_DATA file and return namespace
    set quiet=True to disable verbose output
    """

    # open file
    with open(file, 'r') as f:

        # read in header
        while True:
            line = f.readline()
            linearr = line.split()
            if 'Format date' in line:
                date = linearr[0]
            elif 'Number of species' in line:
                nspec = int(linearr[0])
            elif 'Total number of isotopes' in line:
                niso = int(linearr[0])
            elif 'Maximum number of isotopes/species' in line:
                maxisospec = int(linearr[0])
            elif 'Number of reactions' in line:
                nreac = int(linearr[0])
                break
        if not quiet:
            print('*** INPUT FILE: ' + file)
            print('Format date: ' + str(date))
            print('Number of species: ' + str(nspec))
            print('Total number of isotopes: ' + str(niso))
            print('Maximum number of isotopes/species: ' + str(maxisospec))
            print('Number of reactions: ' + str(nreac))

        # isotopes
        isospec = []
        amu = np.zeros(niso)
        aiso = np.zeros(niso, dtype='int32')
        stable = []  # True or False
        for i in range(niso):
            line = ''
            while len(line.strip()) == 0 or line[0] == '!':
                line = f.readline()
            linearr = line.split()
            isospec.append(linearr[0])
            amu[i] = float(linearr[1])
            aiso[i] = np.rint(amu[i])
            stable.append(True) if linearr[2] == 's' else stable.append(False)
        if not quiet:
            print("INFO - Read in isotope information")

        # decay chains
        isospec_parent = []
        amu_parent = np.zeros(nreac)
        aiso_parent = np.zeros(nreac, dtype='int32')
        thalf = np.zeros(nreac)
        decay_const = np.zeros(nreac)
        isospec_daughter = []
        amu_daughter = np.zeros(nreac)
        aiso_daughter = np.zeros(nreac, dtype='int32')
        edec = np.zeros(nreac)
        seqnum = []
        nlines = np.zeros(nreac, dtype='int32')

        nchains = 0
        for i in range(nreac):
            line = ''
            while len(line.strip()) == 0 or line[0] == '!':
                line = f.readline()
            linearr = line.split()
            isospec_parent.append(linearr[0])
            amu_parent[i] = float(linearr[1])
            aiso_parent[i] = np.rint(amu_parent[i])
            thalf[i] = float(linearr[2]) * DAY2SEC  # convert to seconds
            decay_const[i] = np.log(2) / thalf[i]
            isospec_daughter.append(linearr[3])
            amu_daughter[i] = float(linearr[4])
            aiso_daughter[i] = np.rint(amu_daughter[i])
            edec[i] = float(linearr[5]) * MEV2ERG  # convert to ergs
            seqnum.append(linearr[6])
            if seqnum[-1] in ['F', 'E']:
                nchains = nchains + 1
            nlines[i] = int(linearr[7])
        if not quiet:
            print("INFO - Read in decay chains")

    # output
    out = {}
    out['date'] = date
    out['nspec'] = nspec
    out['niso'] = niso
    out['maxisospec'] = maxisospec
    out['nreac'] = nreac
    out['nchains'] = nchains
    out['isospec'] = isospec
    out['amu'] = amu
    out['aiso'] = aiso
    out['stable'] = stable
    out['isospec_parent'] = isospec_parent
    out['amu_parent'] = amu_parent
    out['aiso_parent'] = aiso_parent
    out['thalf'] = thalf
    out['decay_const'] = decay_const
    out['isospec_daughter'] = isospec_daughter
    out['amu_daughter'] = amu_daughter
    out['aiso_daughter'] = aiso_daughter
    out['edec'] = edec
    out['seqnum'] = seqnum
    out['nlines'] = nlines

    # end
    return out

###############################################################################


def rd_sn_hydro_data(file, ncol=8, reverse=False,
                     quiet=False):
    """read in SN_HYDRO_DATA or SN_HYDRO_FOR_NEXT_MODEL files and return namespace
    set reverse=True to output vectors from vmin to vmax (CMFGEN's grid moves inward from vmax to vmin)
    set quiet=True to disable verbose output
    """

    MAX_POP_DIFF = 1e-5  # maximum absolute difference between sum(isofrac) and corresponding specfrac

    # open file
    with open(file, 'r') as f:

        # read in header
        okhdr = 0
        nd, nspec, niso = 0, 0, 0
        time = 0.0
        while okhdr == 0:
            line = f.readline()
            if 'Number of data points:' in line:
                nd = int(line.split()[4])
                nrow = int(np.ceil(nd / float(ncol)))
            elif 'Number of mass fractions:' in line:
                nspec = int(line.split()[4])
            elif 'Number of isotopes:' in line:
                niso = int(line.split()[3])
            elif 'Time(days) since explosion:' in line:
                time = float(line.split()[3])
            elif 'Radius grid' in line:
                if nd == 0 or nspec == 0 or niso == 0 or time == 0.0:
                    sys.exit('nd, nspec, niso or model time undefined')
                else:
                    okhdr = 1
        if not quiet:
            print(' *** INPUT FILE: ' + file)
            print(' Number of data points:          ' + str(nd))
            print(' Number of mass fractions:       ' + str(nspec))
            print(' Number of isotopes:             ' + str(niso))
            print(' Time(days) since explosion:     ' + str(time))

        # read in hydro grid vectors
        rad = np.zeros(nd)       # radius grid (10^10 cm)
        vel = np.zeros(nd)       # velocity (km/s)
        sigma = np.zeros(nd)     # sigma = dlnV/dlnR-1 (=0 for pure hubble flow)
        temp = np.zeros(nd)      # temperature (10^4 K)
        dens = np.zeros(nd)      # mass density (g/cm^3)
        atomdens = np.zeros(nd)  # atom density (/cm^3)
        ed = np.zeros(nd)        # electron density (/cm^3)
        rossopac = np.zeros(nd)  # Rossland mean opacity (10^-10 cm^-1)
        kappa = np.zeros(nd)     # mass absorption coefficient (cm^2/g)
        okhydro = 0
        while okhydro == 0:
            while line == '':
                line = f.readline()
            if 'Radius grid' in line:
                rad = np.fromfile(f, count=nd, sep=" ", dtype=float)
            elif 'Velocity' in line:
                vel = np.fromfile(f, count=nd, sep=" ", dtype=float)
            elif 'Sigma' in line:
                sigma = np.fromfile(f, count=nd, sep=" ", dtype=float)
            elif 'Temperature' in line:
                temp = np.fromfile(f, count=nd, sep=" ", dtype=float)
            elif 'Density' in line:
                dens = np.fromfile(f, count=nd, sep=" ", dtype=float)
            elif 'Atom density' in line:
                atomdens = np.fromfile(f, count=nd, sep=" ", dtype=float)
            elif 'Electron density' in line:
                ed = np.fromfile(f, count=nd, sep=" ", dtype=float)
            elif 'Rosseland mean' in line:
                rossopac = np.fromfile(f, count=nd, sep=" ", dtype=float)
            elif 'Kappa' in line:
                kappa = np.fromfile(f, count=nd, sep=" ", dtype=float)
            elif 'mass fraction' in line:
                if rad[0] == 0.0 or temp[0] == 0.0 or atomdens[0] == 0.0 or ed[0] == 0.0:
                    sys.exit('Error reading SN hydro data: R or T is zero')
                else:
                    okhydro = 1
            if okhydro == 0:
                line = ''
        if not quiet:
            print(' INFO - Read in hydro grid vectors')

        # compute shell volumes and masses
        dvol = np.zeros(nd)   # cm^3
        dmass = np.zeros(nd)  # g
        rad_cgs = rad * 1e10  # rad in cm
        dvol[0] = rad_cgs[0]**3 - (0.5 * (rad_cgs[1] + rad_cgs[0]))**3
        dvol[nd - 1] = (0.5 * (rad_cgs[nd - 1] + rad_cgs[nd - 2]))**3 - rad_cgs[nd - 1]**3
        for i in range(1, nd - 1):
            rmin = 0.5 * (rad_cgs[i + 1] + rad_cgs[i])
            rmax = 0.5 * (rad_cgs[i] + rad_cgs[i - 1])
            dvol[i] = rmax**3 - rmin**3
        dvol = dvol * 4.0 / 3.0 * np.pi
        dmass = dens * dvol
        if not quiet:
            print(' INFO - Computed shell volumes and masses')

        # read in mass fractions
        spec = []
        specfrac = np.zeros((nd, nspec))
        for ispec in range(nspec):
            while 'mass fraction' not in line:
                line = f.readline()
            spec.append(line.split()[0])
            for ii in range(nrow):
                line = f.readline()
                if '*0.' in line:
                    specfrac[:, ispec] = 0.0
                    if not quiet:
                        print(' INFO - set mass fraction = 0.0 everywhere for ' + spec[ispec])
                    break
                else:
                    entries = line.split()
                    # set mass fractions to 0.0 if < 1D-99 (written e.g. 1.0000000-100)
                    entries = [entries[k] if 'E' in entries[k] else '0.0' for k in range(len(entries))]
                    idx0 = ii * ncol
                    idx1 = idx0 + len(entries)
                    specfrac[idx0:idx1, ispec] = np.array([float(xx) for xx in entries])
            line = ''
        if not quiet:
            print(' INFO - Read in species mass fractions')

        # read in isotope mass fractions
        iso = []
        aiso = np.zeros(niso, dtype='int32')
        isofrac = np.zeros((nd, niso))
        for iiso in range(niso):
            while 'mass fraction' not in line:
                line = f.readline()
            iso.append(line.split()[0])
            aiso[iiso] = int(line.split()[1])
            for ii in range(nrow):
                line = f.readline()
                if '*0.' in line:
                    isofrac[:, iiso] = 0.0
                    if not quiet:
                        print(' INFO - set mass fraction = 0.0 everywhere for ' + iso[iiso] + ' ' + str(aiso[iiso]))
                    break
                else:
                    entries = line.split()
                    # set mass fractions to 0.0 if < 1D-99 (written e.g. 1.0000000-100)
                    entries = [entries[k] if 'E' in entries[k] else '0.0' for k in range(len(entries))]
                    idx0 = ii * ncol
                    idx1 = idx0 + len(entries)
                    isofrac[idx0:idx1, iiso] = np.array([float(xx) for xx in entries])
            line = ''
        if not quiet:
            print(' INFO - Read in isotope mass fractions')

    # check sum isotope mass fractions = species mass fractions
    for s in list(set(iso)):
        # find all indices of species name in iso list
        idx = [i for i, x in enumerate(iso) if x == s]
        # sum isotope mass fractions
        sumisofrac = np.zeros(nd)
        for iiso in range(len(idx)):
            sumisofrac += isofrac[:, idx[iiso]]
        # compare to corresponding species mass fraction
        absdiff = np.abs(specfrac[:, spec.index(s)] - sumisofrac)
        relabsdiff = absdiff / sumisofrac
        if np.max(relabsdiff) > MAX_POP_DIFF:
            sys.exit("ERROR - Maximum absolute difference > MAX_POP_DIFF for species {0:s}".format(s))

    # reversed vectors if reverse=True
    if reverse:
        rad = rad[::-1]
        vel = vel[::-1]
        sigma = sigma[::-1]
        temp = temp[::-1]
        dens = dens[::-1]
        atomdens = atomdens[::-1]
        ed = ed[::-1]
        rossopac = rossopac[::-1]
        kappa = kappa[::-1]
        for i in range(nspec):
            specfrac[:, i] = specfrac[::-1, i]
        for i in range(niso):
            isofrac[:, i] = isofrac[::-1, i]
        dvol = dvol[::-1]
        dmass = dmass[::-1]

    # output
    out = {}
    out['nd'] = nd
    out['nspec'] = nspec
    out['niso'] = niso
    out['time'] = time
    out['rad'] = rad
    out['vel'] = vel
    out['sigma'] = sigma
    out['temp'] = temp
    out['dens'] = dens
    out['atomdens'] = atomdens
    out['ed'] = ed
    out['rossopac'] = rossopac
    out['kappa'] = kappa
    out['spec'] = spec
    out['specfrac'] = specfrac
    out['iso'] = iso
    out['aiso'] = aiso
    out['isofrac'] = isofrac
    out['dvol'] = dvol
    out['dmass'] = dmass

    # end
    return out

###############################################################################


if __name__ == '__main__':

    a = rd_sn_hydro_data('DDC10/SN_HYDRO_DATA_0.976d')
    b = rd_nuc_decay_data('NUC_DECAY_DATA')
    print(a['vel'])
    print(b['amu_parent'])



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK

import argcomplete
import argparse

import numpy as np
# import io
import pandas as pd
# import math
import multiprocessing
from functools import partial
from pathlib import Path
import matplotlib.pyplot as plt

import artistools as at
# import artistools.estimators


def plot_qdot(modelpath, dfpartcontrib, dfmodel, allparticledata, arr_time_artis_days, arr_time_gsi_days, pdfoutpath):
    try:
        depdata = at.get_deposition(modelpath=modelpath)
    except FileNotFoundError:
        print("Can't do qdot plot because no deposition.out file")
        return

    modelname = at.get_model_name(modelpath)

    tstart = depdata['tmid_days'].min()
    tend = depdata['tmid_days'].max()

    arr_heat = {}

    heatcols = ['hbeta', 'halpha', 'hbfis', 'hspof']  # , 'Qdot'

    for col in heatcols:
        arr_heat[col] = np.zeros_like(arr_time_gsi_days)

    modelmass = dfmodel.cellmass_grams.sum()
    print("model mass: {modelmass} g")
    dfpartcontrib = dfpartcontrib.query('particleid in @allparticledata.keys()')
    for cellindex, dfpartcontrib in dfpartcontrib.groupby('cellindex'):
        if cellindex >= len(dfmodel):
            continue
        cell_mass_frac = dfmodel.iloc[cellindex - 1].cellmass_grams / modelmass
        if cell_mass_frac == 0.:
            continue
        frac_of_cellmass_sum = dfpartcontrib.frac_of_cellmass.sum()

        for particleid, frac_of_cellmass in dfpartcontrib[['particleid', 'frac_of_cellmass']].itertuples(index=False):
            thisparticledata = allparticledata[particleid]
            for col in heatcols:
                arr_heat[col] += (
                    thisparticledata[col] * cell_mass_frac * frac_of_cellmass / frac_of_cellmass_sum)

    fig, axis = plt.subplots(
        nrows=1, ncols=1, sharex=False, sharey=False, figsize=(6, 4),
        tight_layout={"pad": 0.4, "w_pad": 0.0, "h_pad": 0.0})

    axis.set_ylim(bottom=1e7, top=2e10)
    axis.set_xlim(left=tstart, right=tend)

    # axis.set_xscale('log')

    # axis.set_xlim(left=1., right=arr_time_artis[-1])
    axis.set_xlabel('Time [days]')
    axis.set_yscale('log')
    # axis.set_ylabel(f'X({strnuc})')
    axis.set_ylabel('Qdot [erg/g/s]')
    # arr_time_days, arr_qdot = zip(
    #     *[(t, qdot) for t, qdot in zip(arr_time_days, arr_qdot)
    #       if depdata['tmid_days'].min() <= t and t <= depdata['tmid_days'].max()])

    # axis.plot(arr_time_gsi_days, arr_heat['Qdot'],
    #           # linestyle='None',
    #           linewidth=2, color='black',
    #           # marker='x', markersize=8,
    #           label='Qdot GSI Network')
    #
    # axis.plot(depdata['tmid_days'], depdata['Qdot_ana_erg/g/s'],
    #           linewidth=2, color='red',
    #           # linestyle='None',
    #           # marker='+', markersize=15,
    #           label='Qdot ARTIS')

    axis.plot(arr_time_gsi_days, arr_heat['hbeta'],
              linewidth=2, color='black',
              linestyle='dashed',
              # marker='x', markersize=8,
              label=r'$\dot{Q}_\beta$ GSI Network')

    axis.plot(depdata['tmid_days'], depdata['Qdot_betaminus_ana_erg/g/s'],
              linewidth=2, color='red',
              linestyle='dashed',
              # marker='+', markersize=15,
              label=r'$\dot{Q}_\beta$ ARTIS')

    axis.plot(arr_time_gsi_days, arr_heat['halpha'],
              linewidth=2, color='black',
              linestyle='dotted',
              # marker='x', markersize=8,
              label=r'$\dot{Q}_\alpha$ GSI Network')

    axis.plot(depdata['tmid_days'], depdata['Qdotalpha_ana_erg/g/s'],
              linewidth=2, color='red',
              linestyle='dotted',
              # marker='+', markersize=15,
              label=r'$\dot{Q}_\alpha$ ARTIS')

    axis.plot(arr_time_gsi_days, arr_heat['hbfis'],
              linewidth=2,
              linestyle='dotted',
              # marker='x', markersize=8,
              # color='black',
              label=r'$\dot{Q}_{hbfis}$ GSI Network')

    axis.plot(arr_time_gsi_days, arr_heat['hspof'],
              linewidth=2,
              linestyle='dotted',
              # marker='x', markersize=8,
              # color='black',
              label=r'$\dot{Q}_{spof}$ GSI Network')

    axis.legend(loc='best', frameon=False, handlelength=1, ncol=3,
                numpoints=1)

    fig.suptitle(f'{modelname}', fontsize=10)
    plt.savefig(pdfoutpath, format='pdf')
    print(f'Saved {pdfoutpath}')


def plot_abund(
        modelpath, dfpartcontrib, allparticledata, arr_time_artis_days, arr_time_gsi_days,
        arr_strnuc, arr_abund_gsi, arr_abund_artis, t_model_init_days, dfcell, pdfoutpath, mgi):

    dfpartcontrib_thiscell = dfpartcontrib.query('cellindex == (@mgi + 1) and particleid in @allparticledata.keys()')
    frac_of_cellmass_sum = dfpartcontrib_thiscell.frac_of_cellmass.sum()
    print(f'frac_of_cellmass_sum: {frac_of_cellmass_sum} (can be < 1.0 because of missing particles)')

    for strnuc in arr_strnuc:
        arr_abund_gsi[strnuc] = np.zeros_like(arr_time_gsi_days)

    for particleid, frac_of_cellmass in dfpartcontrib_thiscell[
            ['particleid', 'frac_of_cellmass']].itertuples(index=False):
        frac_of_cellmass = dfpartcontrib_thiscell.query('particleid == @particleid').frac_of_cellmass.sum()

        for strnuc in arr_strnuc:
            arr_abund_gsi[strnuc] += (
                allparticledata[particleid][strnuc] * frac_of_cellmass / frac_of_cellmass_sum)

    fig, axes = plt.subplots(
        nrows=len(arr_strnuc), ncols=1, sharex=False, sharey=False, figsize=(6, 1 + 2. * len(arr_strnuc)),
        tight_layout={"pad": 0.4, "w_pad": 0.0, "h_pad": 0.0})
    fig.subplots_adjust(top=0.8)
    # axis.set_xscale('log')

    modelname = at.get_model_name(modelpath)

    axes[-1].set_xlabel('Time [days]')
    for axis, strnuc in zip(axes, arr_strnuc):
        # print(arr_time_artis_days)
        xmin = arr_time_gsi_days.min() * 0.9
        xmax = arr_time_gsi_days.max() * 1.03
        axis.set_xlim(left=xmin, right=xmax)
        # axis.set_yscale('log')
        # axis.set_ylabel(f'X({strnuc})')
        axis.set_ylabel('Mass fraction')

        axis.plot(arr_time_gsi_days, arr_abund_gsi[strnuc],
                  # linestyle='None',
                  linewidth=2,
                  marker='x', markersize=8,
                  label=f'{strnuc} GSI Network', color='black')

        if strnuc in arr_abund_artis:
            axis.plot(arr_time_artis_days, arr_abund_artis[strnuc],
                      linewidth=2,
                      # linestyle='None',
                      # marker='+', markersize=15,
                      label=f'{strnuc} ARTIS', color='red')

        if f'X_{strnuc}' in dfcell:
            axis.plot(t_model_init_days, dfcell[f'X_{strnuc}'],
                      marker='+', markersize=15, markeredgewidth=2,
                      label=f'{strnuc} ARTIS inputmodel', color='blue')
            print(strnuc, arr_abund_gsi[strnuc][0], dfcell[f'X_{strnuc}'])

        axis.legend(loc='best', frameon=False, handlelength=1, ncol=1, numpoints=1)

    fig.suptitle(f'{modelname} cell {mgi}', y=0.995, fontsize=10)
    plt.savefig(pdfoutpath, format='pdf')
    print(f'Saved {pdfoutpath}')


def get_particledata(arr_time_s, arr_strnuc, particleid):

    try:
        nts_min = at.inputmodel.rprocess_from_trajectory.get_closest_network_timestep(
            particleid, timesec=min(arr_time_s), cond='lessthan')
        nts_max = at.inputmodel.rprocess_from_trajectory.get_closest_network_timestep(
            particleid, timesec=max(arr_time_s), cond='greaterthan')

    except FileNotFoundError:
        # print(f'WARNING: Particle data not found for id {particleid}')
        return 'NONE', None

    # print(f'Reading data for particle id {particleid}...')
    particledata = {
        'Qdot': {},
        'hbeta': {},
        'halpha': {},
        'hbfis': {},
        'hspof': {},
        **{strnuc: {} for strnuc in arr_strnuc}
    }
    nstep_timesec = {}
    with at.inputmodel.rprocess_from_trajectory.open_tar_file_or_extracted(
            particleid, './Run_rprocess/heating.dat') as f:

        dfheating = pd.read_csv(
            f, delim_whitespace=True, usecols=[
                '#count', 'time/s', 'hbeta', 'halpha', 'hbfis', 'hspof'])
        heatcols = ['hbeta', 'halpha', 'hbfis', 'hspof']

        heatrates_in = {col: [] for col in heatcols}
        arr_time_s_source = []
        for _, row in dfheating.iterrows():
            nstep_timesec[row['#count']] = row['time/s']
            arr_time_s_source.append(row['time/s'])
            for col in heatcols:
                try:
                    heatrates_in[col].append(float(row[col]))
                except ValueError:
                    heatrates_in[col].append(float(row[col].replace('-', 'e-')))

        for col in heatcols:
            particledata[col] = np.interp(arr_time_s, arr_time_s_source, heatrates_in[col])

    with at.inputmodel.rprocess_from_trajectory.open_tar_file_or_extracted(
            particleid, './Run_rprocess/energy_thermo.dat') as f:
        dfthermo = pd.read_csv(f, delim_whitespace=True, usecols=['#count', 'time/s', 'Qdot'])

        heatcols = ['Qdot']
        heatrates_in = {col: [] for col in heatcols}
        arr_time_s_source = []
        for _, row in dfthermo.iterrows():
            nstep_timesec[row['#count']] = row['time/s']
            arr_time_s_source.append(row['time/s'])
            for col in heatcols:
                try:
                    heatrates_in[col].append(float(row[col]))
                except ValueError:
                    heatrates_in[col].append(float(row[col].replace('-', 'e-')))

        for col in heatcols:
            particledata[col] = np.interp(arr_time_s, arr_time_s_source, heatrates_in[col])

    if arr_strnuc:
        arr_traj_time_s = []
        arr_massfracs = {strnuc: [] for strnuc in arr_strnuc}
        for nts in range(nts_min, nts_max + 1):
            timesec = nstep_timesec[nts]
            arr_traj_time_s.append(timesec)
            # print(nts, timesec / 86400)
            traj_nuc_abund = at.inputmodel.rprocess_from_trajectory.get_trajectory_nuc_abund(particleid, nts=nts)
            for strnuc in arr_strnuc:
                arr_massfracs[strnuc].append(traj_nuc_abund.get(f'X_{strnuc}', 0.))

        for strnuc in arr_strnuc:
            massfracs_interp = np.interp(arr_time_s, arr_traj_time_s, arr_massfracs[strnuc])
            particledata[strnuc] = massfracs_interp

    return particleid, particledata


def do_modelcells(modelpath, mgiplotlist, arr_el_a):
    arr_el, arr_a = zip(*arr_el_a)
    arr_strnuc = [z + str(a) for z, a in arr_el_a]

    # arr_z = [at.get_atomic_number(el) for el in arr_el]

    dfmodel, t_model_init_days, vmax_cmps = at.inputmodel.get_modeldata(modelpath)
    npts_model = len(dfmodel)
    MH = 1.67352e-24  # g

    arr_time_artis_days = []
    arr_abund_artis = {}

    try:
        estimators = at.estimators.read_estimators(modelpath, modelgridindex=tuple(mgiplotlist))
        for nts, mgi in estimators.keys():
            if mgi not in mgiplotlist:
                continue

            time_days = float(estimators[(nts, mgi)]['tdays'])
            if mgi == mgiplotlist[0]:
                arr_time_artis_days.append(time_days)

            rho_init_cgs = 10 ** dfmodel.iloc[mgi].logrho
            rho_cgs = rho_init_cgs * (t_model_init_days / time_days) ** 3
            for strnuc, a in zip(arr_strnuc, arr_a):
                abund = estimators[(nts, mgi)]['populations'].get(strnuc, 0.)
                massfrac = abund * a * MH / rho_cgs

                if mgi not in arr_abund_artis:
                    arr_abund_artis[mgi] = {}

                if strnuc not in arr_abund_artis[mgi]:
                    arr_abund_artis[mgi][strnuc] = []

                arr_abund_artis[mgi][strnuc].append(massfrac)

    except FileNotFoundError:
        pass

    arr_time_artis_s = np.array([t * 8.640000E+04 for t in arr_time_artis_days])

    arr_abund_gsi = {}

    arr_time_gsi_s = np.array([t_model_init_days * 86400, *arr_time_artis_s])
    arr_time_gsi_days = arr_time_gsi_s / 86400

    dfpartcontrib = at.inputmodel.rprocess_from_trajectory.get_gridparticlecontributions(modelpath)
    dfpartcontrib.query('cellindex <= @npts_model', inplace=True)

    list_particleids_getabund = dfpartcontrib.query('(cellindex - 1) in @mgiplotlist').particleid.unique()
    fworkerwithabund = partial(get_particledata, arr_time_gsi_s, arr_strnuc)

    print(f'Reading trajectory data for {len(list_particleids_getabund)} particles with abundances')

    if at.num_processes > 1:
        with multiprocessing.Pool(processes=at.num_processes) as pool:
            list_particledata_withabund = pool.map(fworkerwithabund, list_particleids_getabund)
            pool.close()
            pool.join()
    else:
        list_particledata_withabund = [fworkerwithabund(particleid) for particleid in list_particleids_getabund]

    list_particleids_noabund = [
        pid for pid in dfpartcontrib.particleid.unique() if pid not in list_particleids_getabund]
    fworkernoabund = partial(get_particledata, arr_time_gsi_s, [])
    print(f'Reading trajectory data for {len(list_particleids_noabund)} particles for Qdot/thermal data (no abundances)')

    if at.num_processes > 1:
        with multiprocessing.Pool(processes=at.num_processes) as pool:
            list_particledata_noabund = pool.map(fworkernoabund, list_particleids_noabund)
            pool.close()
            pool.join()
    else:
        list_particledata_noabund = [fworkernoabund(particleid) for particleid in list_particleids_noabund]

    allparticledata = {
        particleid: data for particleid, data in (list_particledata_withabund + list_particledata_noabund)}

    plot_qdot(
        modelpath, dfpartcontrib, dfmodel, allparticledata, arr_time_artis_days, arr_time_gsi_days,
        pdfoutpath=Path(modelpath, 'gsinetwork_global-qdot.pdf'))

    for mgi in mgiplotlist:
        plot_abund(
            modelpath, dfpartcontrib, allparticledata, arr_time_artis_days, arr_time_gsi_days, arr_strnuc,
            arr_abund_gsi, arr_abund_artis.get(mgi, []), t_model_init_days, dfmodel.iloc[mgi], mgi=mgi,
            pdfoutpath=Path(modelpath, f'gsinetwork_cell{mgi}-abundance.pdf'))


def addargs(parser):
    parser.add_argument('-modelpath',
                        default='.',
                        help='Path for ARTIS files')

    parser.add_argument('-outputpath', '-o',
                        default='.',
                        help='Path for output files')

    parser.add_argument('-modelgridindex', '-cell', '-mgi', default=0,
                        help='Modelgridindex (zero-indexed) to plot or list such as 4,5,6')


def main(args=None, argsraw=None, **kwargs):
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Create solar r-process pattern in ARTIS format.')

        addargs(parser)
        parser.set_defaults(**kwargs)
        argcomplete.autocomplete(parser)
        args = parser.parse_args(argsraw)

    arr_el_a = [
        ('Ga', 72),
        ('Sr', 89),
        ('Ba', 140),
        # ('Ce', 141),
        ('Nd', 147),
        ('Rn', 222),
        ('Ra', 223),
        ('Ra', 224),
        ('Ra', 225),
        ('Ac', 225),
        ('Th', 234),
        ('Pa', 233),
        ('U', 235),
    ]
    arr_el_a.sort(key=lambda x: (at.elsymbols.index(x[0]), -x[1]))

    modelpath = Path(args.modelpath)
    if hasattr(args.modelgridindex, 'split'):
        mgiplotlist = [int(mgi) for mgi in args.modelgridindex.split(',')]
    else:
        mgiplotlist = [int(args.modelgridindex)]

    do_modelcells(modelpath, mgiplotlist, arr_el_a)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import glob
import multiprocessing
import sys

import matplotlib.pyplot as plt

xminvalue, xmaxvalue = 3500, 7000
# xminvalue, xmaxvalue = 10000, 20000

h = 6.62607004e-34  # m^2 kg / s
c = 299792458  # m / s


def main():
    logfiles = glob.glob('output_0-0.txt') + glob.glob('*/output_0-0.txt') + \
        glob.glob('*/*/output_0-0.txt')
    if not logfiles:
        print('no output log files found')
        sys.exit()

    fig, axes = plt.subplots(
        nrows=2, ncols=1, sharey=True, figsize=(8, 5), tight_layout={"pad": 0.2, "w_pad": 0.0, "h_pad": 0.0})

    for index, logfilename in enumerate(logfiles):
        runfolder = logfilename.split('/output_0-0.txt')[0]

        timesteptimes = []
        with open(runfolder + '/light_curve.out', 'r') as lcfile:
            for line in lcfile:
                timesteptimes.append(line.split()[0])

        timesteptimes = timesteptimes[:int(len(timesteptimes) / 2)]

        stats = []

        with open(logfilename) as flog:
            for line in flog:
                if line.startswith('timestep '):
                    currenttimestep = int(line.split(' ')[1].split(',')[0])
                    stats.append({})
                    if len(stats) != currenttimestep + 1:
                        print("WRONG TIMESTEP!")
                if line.startswith('k_stat_'):
                    (key, value) = line.split(' = ')
                    stats[-1][key] = int(value)

        linelabel = runfolder

        linestyle = ['-', '--'][int(index / 7)]
        # xvalues = range(len(stats))
        yvalues = [timestepstats['k_stat_to_r_fb'] for timestepstats in stats]
        axes[0].plot(timesteptimes, yvalues, linestyle=linestyle, linewidth=1.5, label=linelabel)
        yvalues = [timestepstats['k_stat_to_ma_collexc'] for timestepstats in stats]
        axes[1].plot(timesteptimes, yvalues, linestyle=linestyle, linewidth=1.5, label=linelabel)

    for axis in axes:
        axis.set_xlim(250, 300)
        # axis.set_ylim(-0.1, 1.3)

    axes[0].legend(loc='best', handlelength=2, frameon=False, numpoints=1, prop={'size': 9})
    axes[-1].set_xlabel(r'Time (days)')
    # ax.xaxis.set_minor_locator(ticker.MultipleLocator(base=5))
    axes[0].set_ylabel(r'k_stat_to_r_fb')
    axes[1].set_ylabel(r'k_stat_to_ma_collexc')

    fig.savefig('plotartisstats.pdf', format='pdf')
    plt.close()


if __name__ == "__main__":
    multiprocessing.freeze_support()
    main()



#!/usr/bin/env python3
import argparse
import glob
import multiprocessing
import os.path
import sys

from astropy import constants as const
# from astropy import units as u

import artistools as at
import matplotlib.pyplot as plt
# import matplotlib.ticker as ticker
import pandas as pd

defaultoutputfile = 'plotmacroatom_cell{0:03d}_{1:03d}-{2:03d}.pdf'


def addargs(parser):
    parser.add_argument('--modelpath', nargs='?', default='',
                        help='Path to ARTIS folder')
    parser.add_argument('-timestep', type=int, default=10,
                        help='Timestep number to plot, or -1 for last')
    parser.add_argument('-timestepmax', type=int, default=-1,
                        help='Make plots for all timesteps up to this timestep')
    parser.add_argument('-modelgridindex', '-cell', type=int, default=0,
                        help='Modelgridindex to plot')
    parser.add_argument('element', nargs='?', default='Fe',
                        help='Plotted element')
    parser.add_argument('-xmin', type=int, default=1000,
                        help='Plot range: minimum wavelength in Angstroms')
    parser.add_argument('-xmax', type=int, default=15000,
                        help='Plot range: maximum wavelength in Angstroms')
    parser.add_argument('-o', action='store', dest='outputfile',
                        default=defaultoutputfile,
                        help='Filename for PDF file')


def main(args=None, argsraw=None, **kwargs):
    """Plot the macroatom transitions."""
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Plot ARTIS macroatom transitions.')
        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

    if os.path.isdir(args.outputfile):
        args.outputfile = os.path.join(args.outputfile, defaultoutputfile)

    try:
        atomic_number = next(Z for Z, elsymb in enumerate(at.elsymbols) if elsymb.lower() == args.element.lower())
    except StopIteration:
        print(f"Could not find element '{args.element}'")
        return

    timestepmin = args.timestep

    if not args.timestepmax or args.timestepmax < 0:
        timestepmax = timestepmin
    else:
        timestepmax = args.timestepmax

    input_files = (
        glob.glob(os.path.join(args.modelpath, 'macroatom_????.out*'), recursive=True) +
        glob.glob(os.path.join(args.modelpath, '*/macroatom_????.out*'), recursive=True))

    if not input_files:
        print("No macroatom files found")
        return 1
    else:
        dfall = read_files(input_files, args.modelgridindex, timestepmin, timestepmax, atomic_number)

    specfilename = os.path.join(args.modelpath, 'spec.out')

    if not os.path.isfile(specfilename):
        specfilename = os.path.join(args.modelpath, '../example_run/spec.out')

    if not os.path.isfile(specfilename):
        print(f'Could not find {specfilename}')
        return 1

    outputfile = args.outputfile.format(args.modelgridindex, timestepmin, timestepmax)
    make_plot(dfall, args.modelpath, specfilename, timestepmin, timestepmax, outputfile,
              xmin=args.xmin, xmax=args.xmax, modelgridindex=args.modelgridindex)

    return 0


def make_plot(dfmacroatom, modelpath, specfilename, timestepmin, timestepmax, outputfile,
              xmin, xmax, modelgridindex, nospec=False, normalised=False):

    time_days_min = at.get_timestep_time(modelpath, timestepmin)
    time_days_max = at.get_timestep_time(modelpath, timestepmax)

    print(f'Plotting {len(dfmacroatom)} transitions')

    fig, axis = plt.subplots(
        nrows=1, ncols=1, sharex=True, figsize=(6, 6), tight_layout={"pad": 0.2, "w_pad": 0.0, "h_pad": 0.0})

    axis.annotate(f'Timestep {timestepmin:d} to {timestepmax:d} (t={time_days_min} to '
                  f'{time_days_max})\nCell {modelgridindex:d}',
                  xy=(0.02, 0.96), xycoords='axes fraction',
                  horizontalalignment='left', verticalalignment='top', fontsize=8)

    lambda_cmf_in = const.c.to('angstrom/s').value / dfmacroatom['nu_cmf_in'].values
    lambda_cmf_out = const.c.to('angstrom/s').value / dfmacroatom['nu_cmf_out'].values
    # axis.scatter(lambda_cmf_in, lambda_cmf_out, s=1, alpha=0.5, edgecolor='none')
    axis.plot(lambda_cmf_in, lambda_cmf_out, linestyle='none', marker='o',  # alpha=0.5,
              markersize=2, markerfacecolor='red', markeredgewidth=0)
    axis.set_xlabel(r'Wavelength in ($\AA$)')
    axis.set_ylabel(r'Wavelength out ($\AA$)')
    # axis.xaxis.set_minor_locator(ticker.MultipleLocator(base=100))
    axis.set_xlim(xmin, xmax)
    axis.set_ylim(xmin, xmax)

    # axis.legend(loc='best', handlelength=2, frameon=False, numpoints=1, prop={'size': 13})

    print(f'Saving to {outputfile:s}')
    fig.savefig(outputfile, format='pdf')
    plt.close()


def read_files(files, modelgridindex=None, timestepmin=None, timestepmax=None, atomic_number=None):
    dfall = None
    if not files:
        print("No files")
    else:
        for _, filepath in enumerate(files):
            print(f'Loading {filepath}...')

            df_thisfile = pd.read_csv(filepath, delim_whitespace=True)
            # df_thisfile[['modelgridindex', 'timestep']].apply(pd.to_numeric)
            if modelgridindex:
                df_thisfile.query('modelgridindex==@modelgridindex', inplace=True)
            if timestepmin is not None:
                df_thisfile.query('timestep>=@timestepmin', inplace=True)
            if timestepmax:
                df_thisfile.query('timestep<=@timestepmax', inplace=True)
            if atomic_number:
                df_thisfile.query('Z==@atomic_number', inplace=True)

            if df_thisfile is not None:
                if len(df_thisfile) > 0:
                    if dfall is None:
                        dfall = df_thisfile.copy()
                    else:
                        dfall = dfall.append(df_thisfile.copy(), ignore_index=True)

        if dfall is None or len(dfall) == 0:
            print("No data found")

    return dfall


if __name__ == "__main__":
    multiprocessing.freeze_support()
    sys.exit(main())



import numpy as np
import pandas as pd
from pathlib import Path


def all_cells_same_opacity(modelpath, ngrid):
    cell_opacities = np.array([0.1] * ngrid)

    with open(Path(modelpath) / 'opacity.txt', 'w') as fopacity:
        fopacity.write(f'{ngrid}\n')

        for cellid, opacity in enumerate(cell_opacities):
            fopacity.write(f'{cellid+1}    {opacity}\n')


def opacity_by_Ye(outputfilepath, griddata):
    """Opacities from Table 1 Tanaka 2020"""
    griddata = pd.DataFrame(griddata)
    print("Getting opacity kappa from Ye")

    cell_opacities = np.zeros_like(griddata['cellYe'])

    for index, Ye in enumerate(griddata['cellYe']):
        if Ye == 0. and griddata['rho'][index] == 0:
            cell_opacities[index] = 0.
        elif Ye <= 0.1:
            cell_opacities[index] = 19.5
        elif Ye <= 0.15:
            cell_opacities[index] = 32.2
        elif Ye <= 0.2:
            cell_opacities[index] = 22.3
        elif Ye <= 0.25:
            cell_opacities[index] = 5.6
        elif Ye <= 0.3:
            cell_opacities[index] = 5.36
        elif Ye <= 0.35:
            cell_opacities[index] = 3.3
        else:
            cell_opacities[index] = 0.96

    griddata['opacity'] = cell_opacities

    with open(Path(outputfilepath) / 'opacity.txt', 'w') as fopacity:
        fopacity.write(f'{len(griddata["inputcellid"])}\n')
        griddata[['inputcellid', 'opacity']].to_csv(fopacity, sep='\t', index=False, header=False, float_format='%.10f')


def get_opacity_from_file(modelpath):
    opacity_file_contents = np.loadtxt(Path(modelpath) / "opacity.txt", unpack=True, skiprows=1)
    opacity = opacity_file_contents[1]

    return opacity


def write_Ye_file(outputfilepath, griddata):
    with open(Path(outputfilepath) / 'Ye.txt', 'w') as fYe:
        fYe.write(f'{len(griddata["inputcellid"])}\n')
        griddata[['inputcellid', 'cellYe']].to_csv(fYe, sep='\t', index=False, header=False, float_format='%.10f')

    print("Saved Ye.txt")
